without trace
without warning

include wildcard.e
include get.e
include graphics.e
include image.e
include dexcept.e 

-- Constants:
global constant LEFT=-1
global constant CENTER=0
global constant RIGHT=1
global constant HARD_SPACE=255
global constant HARD_TAB=repeat(HARD_SPACE,5) 

-- Utilities:
global procedure delay(atom wait)
   atom t1,t2
   t1=time()            
   t2=t1
   while t2-t1<=wait do
      if t2<t1 then t1-=86400 end if
      t2=time()
   end while
end procedure

type is_printable_string(object x)
   integer len
   object char 
   while TRUE do
      if atom(x) then exit end if
      len=length(x)
      for i=1 to len do
         char=x[i]
         if not integer(char) then exit end if
         if char<32 then exit end if
         if char>255 then exit end if 
         if i=len then return TRUE end if
      end for 
      exit
   end while     
   throw(Type_Check_Failure)                          
   return TRUE
end type

type is_extended_string(object x)
   integer len
   object char 
   while TRUE do
      if atom(x) then exit end if
      len=length(x)
      for i=1 to len do
         char=x[i]
         if not integer(char) then exit end if
         if char<32 then
            if not find(char,"\t\n") then exit end if
         end if                                        
         if char>255 then exit end if
         if i=len then return TRUE end if
      end for
      exit
   end while
   throw(Type_Check_Failure)                                
   return TRUE
end type  

type is_foreground(object x)
   if integer(x) then
      if x>=BLACK and x<=BRIGHT_WHITE+BLINKING then return TRUE end if
   end if
   throw(Type_Check_Failure)   
   return TRUE
end type  

type is_background(object x)
   if integer(x) then
      if x>=BLACK and x<=WHITE then return TRUE end if
   end if
   throw(Type_Check_Failure)   
   return TRUE
end type 

type is_line(object x)
   if integer(x) then
      if x>=1 and x<=25 then return TRUE end if
   end if                              
   throw(Type_Check_Failure)   
   return TRUE
end type  

type is_column(object x)
   if integer(x) then
      if x>=1 and x<=80 then return TRUE end if
   end if                              
   throw(Type_Check_Failure)   
   return TRUE
end type

constant ALIGNMENT={LEFT,CENTER,RIGHT}                                              

type is_alignment(object x)
   if find(x,ALIGNMENT) then return TRUE end if
   throw(Type_Check_Failure)   
   return TRUE
end type  

type is_input_limit(object x)
   if integer(x) then
      if x>=0 and x<=78 then return TRUE end if
   end if                                
   throw(Type_Check_Failure)   
   return TRUE
end type


-- Classes: 
package("Diamond Text-Based User Interface")

entity DefaultColor
    
global constant Color=class("Color",FINAL,Entity,NONE)

global type is_Color(object x)
   return live_instance_check(x,Color)    
end type        

property("Foreground",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,BRIGHT_WHITE)
property("Background",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,BLACK)

function Color_new_2(is_foreground foreground,is_background background) 
   entity new            
   if not success() then return this() end if
   new=call_method(this(),"new",NONE)
   set_property(new,"Foreground",foreground)
   set_property(new,"Background",background)
   return new    
end function    

method("new",0,CLASS,FINAL,PUBLIC,SUPER_METHOD,NONE)
method("new",1,CLASS,FINAL,PUBLIC,routine_id("Color_new_2"),{BLACK})
method("new",2,CLASS,FINAL,PUBLIC,routine_id("Color_new_2"),NONE)

method("clone",0,INSTANCE,FINAL,PUBLIC,SUPER_METHOD,NONE)

method("delete",0,INSTANCE,FINAL,PUBLIC,SUPER_METHOD,NONE)

function Color_Set_1(is_Color color)
   if not success() then return this() end if
   set_property(this(),"Foreground",get_property(color,"Foreground"))
   set_property(this(),"Background",get_property(color,"Background"))
   return this()
end function

method("Set",1,INSTANCE,FINAL,PUBLIC,routine_id("Color_Set_1"),NONE)

function Color_Set_2(is_integer foreground,is_integer background)
   integer fg
   if not success() then return this() end if     
   fg=get_property(this(),"Foreground")
   if foreground!=-1 then 
      VOID=is_foreground(foreground)
      if not success() then return this() end if
      set_property(this(),"Foreground",foreground)
   end if
   if background!=-1 then 
      VOID=is_background(background)
      if not success() then
         set_property(this(),"Foreground",fg) 
         return this() 
      end if
      set_property(this(),"Background",background)
   end if                          
   return this()
end function    

method("Set",2,INSTANCE,FINAL,PUBLIC,routine_id("Color_Set_2"),NONE)

function Color_Set_Foreground_1(object color)
   integer background
   background=get_property(this(),"Background")
   VOID=call_method(this(),"Set",{color})
   if success() then
      set_property(this(),"Background",background)   
   end if                                      
   return this()
end function    

method("Set_Foreground",1,INSTANCE,FINAL,PUBLIC,
routine_id("Color_Set_Foreground_1"),NONE)

function Color_Set_Background_1(object color)
   integer foreground
   foreground=get_property(this(),"Foreground")
   VOID=call_method(this(),"Set",{color})
   if success() then
      set_property(this(),"Foreground",foreground)   
   end if                                      
   return this()
end function    

method("Set_Background",1,INSTANCE,FINAL,PUBLIC,
routine_id("Color_Set_Background_1"),NONE)

function Color_Get_Foreground_0()
   return get_property(this(),"Foreground")
end function

method("Get_Foreground",0,INSTANCE,FINAL,PUBLIC,
routine_id("Color_Get_Foreground_0"),NONE) 

function Color_Get_Background_0()
   return get_property(this(),"Background")
end function

method("Get_Background",0,INSTANCE,FINAL,PUBLIC,
routine_id("Color_Get_Background_0"),NONE)                  

function Color_Reverse_0()
   integer foreground,background,modifier
   foreground=get_property(this(),"Foreground")
   background=get_property(this(),"Background")
   modifier=floor(foreground/8)*8
   set_property(this(),"Foreground",background+modifier)
   set_property(this(),"Background",remainder(foreground,8))
   return this()
end function   

method("Reverse",0,INSTANCE,FINAL,PUBLIC,routine_id("Color_Reverse_0"),NONE) 

function Color_Initialize_0()
   DefaultColor=call_method(Color,"new",{BRIGHT_WHITE,BLACK})
   return NIL   
end function

method("Initialize",0,CLASS,FINAL,PRIVATE,
routine_id("Color_Initialize_0"),NONE)

end_class()                    

global constant DEFAULT_COLOR=DefaultColor                             

                                
global constant Screen=class("Screen",FINAL,Entity,NONE)

-- No types are defined as Screen is a unique enitity with no instances
-- and no subclasses.

property("Color",CLASS,PRIVATE,PRIVATE,PERSISTENT,DEFAULT_COLOR)

function Screen_Initialize_0()
   if platform()=DOS32 then VOID=graphics_mode(3) end if
   wrap(FALSE)
   VOID=call_method(Screen,"Clear",NONE)
   return NIL
end function

method("Initialize",0,CLASS,FINAL,PRIVATE,
routine_id("Screen_Initialize_0"),NONE)

function Screen_Set_Color_1(is_Color color)
   if success() then 
      set_property(Screen,"Color",color)
   end if
   return NIL
end function

method("Set_Color",1,CLASS,FINAL,PUBLIC,
routine_id("Screen_Set_Color_1"),NONE)

function Screen_Get_Color_0()
   return get_property(Screen,"Color")
end function

method("Get_Color",0,CLASS,FINAL,PUBLIC,
routine_id("Screen_Get_Color_0"),NONE)

function Screen_Clear_0()
   entity color
   color=get_property(this(),"Color")
   text_color(call_method(color,"Get_Foreground",NONE))
   bk_color(call_method(color,"Get_Background",NONE))
   clear_screen()
   for i=1 to 25 do
      puts(1,repeat(' ',80))
   end for
   position(1,1)
   return NIL
end function

method("Clear",0,CLASS,FINAL,PUBLIC,routine_id("Screen_Clear_0"),NONE) 

function Screen_Print_At_4(is_printable_string text,is_line line,
is_column col,object color)
   entity screenColor
   sequence savePos        
   if not success() then return NIL end if
   if length(text)+col>81 then 
      throw(Type_Check_Failure)
      return NIL
   end if
   screenColor=get_property(Screen,"Color")
   if equal(color,Screen) then
      color=screenColor
   else
      VOID=is_Color(color)
      if not success() then return NIL end if
   end if
   savePos=get_position()
   position(line,col)
   text_color(call_method(color,"Get_Foreground",NONE))
   bk_color(call_method(color,"Get_Background",NONE))
   puts(1,text)
   text_color(call_method(screenColor,"Get_Foreground",NONE))
   bk_color(call_method(screenColor,"Get_Background",NONE))
   position(savePos[1],savePos[2])
   return NIL
end function 

method("Print_At",3,CLASS,FINAL,PUBLIC,
routine_id("Screen_Print_At_4"),{Screen})
method("Print_At",4,CLASS,FINAL,PUBLIC,
routine_id("Screen_Print_At_4"),NONE)

function Screen_Print_Line_3(is_printable_string text,is_alignment alignment,
object color)
   integer width
   sequence pos,spaces
   entity screenColor
   if not success() then return NIL end if
   width=length(text)
   if width>80 then 
      throw(Type_Check_Failure)
      return NIL
   end if
   screenColor=get_property(Screen,"Color") 
   if equal(color,Screen) then
      color=screenColor
   else
      VOID=is_Color(color)
      if not success() then return NIL end if
   end if
   spaces={0,0}
   if alignment=LEFT then
      spaces[1]=80-width
   elsif alignment=CENTER then
      spaces[1]=floor((80-width)/2)
      spaces[2]=floor((81-width)/2)
   else
      spaces[2]=80-width
   end if
   pos=get_position()
   if pos[2]!=1 then puts(1,'\n') end if 
   puts(1,repeat(' ',spaces[1])) 
   text_color(call_method(color,"Get_Foreground",NONE))
   bk_color(call_method(color,"Get_Background",NONE))
   puts(1,text)
   text_color(call_method(screenColor,"Get_Foreground",NONE))
   bk_color(call_method(screenColor,"Get_Background",NONE))
   puts(1,repeat(' ',spaces[2]))
   return NIL   
end function

method("Print_Line",1,CLASS,FINAL,PUBLIC,
routine_id("Screen_Print_Line_3"),{LEFT,Screen})
method("Print_Line",2,CLASS,FINAL,PUBLIC,
routine_id("Screen_Print_Line_3"),{Screen})
method("Print_Line",3,CLASS,FINAL,PUBLIC,
routine_id("Screen_Print_Line_3"),NONE)

end_class()


global constant Box=class("Box",ABSTRACT,Entity,NONE)

global type is_Box(object x)
   return live_instance_check(x,Box)   
end type          

property("Text",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)
property("Color",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,DEFAULT_COLOR)

function Box_Set_Text_1(is_extended_string text)
   if success() then
      set_property(this(),"Text",text)
   end if
   return NIL
end function

method("Set_Text",1,INSTANCE,NORMAL,PUBLIC,routine_id("Box_Set_Text_1"),NONE)

function Box_Get_Text_0()
   return get_property(this(),"Text")
end function

method("Get_Text",0,INSTANCE,NORMAL,PUBLIC,routine_id("Box_Get_Text_0"),NONE)

function Box_Set_Color_1(is_Color color)          
   if success() then 
      set_property(this(),"Color",color)   
   end if  
   return NIL
end function

method("Set_Color",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Box_Set_Color_1"),{DEFAULT_COLOR})
method("Set_Color",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Box_Set_Color_1"),NONE)

function Box_Get_Color_0()
   return get_property(this(),"Color")
end function

method("Get_Color",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Box_Get_Color_0"),NONE)

method("Display",0,INSTANCE,ABSTRACT,PROTECTED,ABSTRACT_METHOD,NONE)

end_class() 


                              
global constant Text_Box=class("Text_Box",NORMAL,Box,NONE)

global type is_Text_Box(object x)
   return live_instance_check(x,Text_Box)   
end type          

method("new",0,CLASS,NORMAL,PUBLIC,SUPER_METHOD,NONE)

method("clone",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE)

method("delete",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE)

function Text_Box_Display_0()
   sequence Text,formatText,color
   integer height,width
   Text=call_method(this(),"Get_Text",NONE)
   formatText={NONE}
   height=1
   for i=1 to length(Text) do
      if Text[i]='\n' then
         formatText=append(Text,NONE)
         height+=1
      elsif Text[i]='\t' then
         formatText[height]&=repeat(' ',5)
      else
         formatText[height]&=Text[i]
      end if
   end for
   height=length(formatText)
   for i=1 to height do
      width=length(formatText[i])
      for j=width to 1 by -1 do
         if formatText[i][j]!=' ' then
            formatText[i]=formatText[i][1..j]
            exit
         end if
         if j=1 then formatText[i]=NONE end if
      end for
   end for
   for i=1 to height do
      width=length(formatText[i])
      for j=1 to width do
         if formatText[i][j]!=' ' then
            formatText[i]=formatText[i][j..width]
            exit
         end if
      end for
   end for
   height=1
   while height<=length(formatText) do
      width=length(formatText[height])
      if width>78 then
         if height=length(formatText) then
            formatText=append(formatText,NONE)
         end if
         for i=78 to 1 by -1 do
            if formatText[height][i]=' ' then
               formatText[height+1]=formatText[height][i+1..width]
               &formatText[height+1]
               formatText[height]=formatText[height][1..i-1]
               for j=i-1 to 1 by -1 do
                  if formatText[height][j]!=' ' then
                     formatText[height]=formatText[height][1..j]
                     exit
                  end if
                  if j=1 then
                     formatText[height]=NONE
                  end if
               end for
               exit
            end if
            if i=1 then
               formatText[height+1]=formatText[height][79..width]
               &formatText[height+1]
               formatText[height]=formatText[height][1..78]
            end if
         end for
      end if
      height+=1
   end while
   for i=1 to length(formatText) do
      if not equal(formatText[i],NONE) then
         formatText=formatText[i..length(formatText)]
         exit
      end if
      if i=length(formatText) then formatText=NONE end if
   end for
   for i=length(formatText) to 1 by -1 do
      if not equal(formatText[i],NONE) then
         formatText=formatText[1..i]
         exit
      end if
   end for
   while TRUE do
      height=length(formatText)
      if height>23 then
         Text=formatText[22..length(formatText)]
         formatText=formatText[1..21]&{"*** Press any key for more ***"}
         height=21
      else
         Text=NONE
      end if 
      puts(1,Text&"\n")
      width=0
      for i=1 to length(formatText) do
         if length(formatText[i])>width then
            width=length(formatText[i])
         end if
      end for
      color=call_method(this(),"Get_Color",NONE)
      VOID=call_method(Screen,"Print_Line",{201&repeat(205,width)&187,
      CENTER,color})
      for i=1 to height do
         VOID=call_method(Screen,"Print_Line",
         {186&formatText[i]&repeat(' ',width-length(formatText[i]))
         &186,CENTER,color})
      end for
      if length(Text)>0 then
         VOID=call_method(Screen,"Print_Line",
         {204&repeat(205,width)&185,CENTER,color})
         if width>30 then
            formatText[22]=repeat(' ',floor((width-30)/2))&formatText[22]
            &repeat(' ',floor((width-29)/2))
         end if
         VOID=call_method(Screen,"Print_Line",
         {186&formatText[22]&186,CENTER,color})
      end if
      VOID=call_method(Screen,"Print_Line",
      {200&repeat(205,width)&188,CENTER,color})
      if length(Text)=0 then exit end if
      formatText=Text
      while get_key()!=-1 do end while
      VOID=wait_key()
      VOID=call_method(Screen,"Clear",NONE)
   end while
   delay(1)
   return NIL
end function

method("Display",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Text_Box_Display_0"),NONE)

end_class()


global constant Pause_Box=class("Pause_Box",NORMAL,Text_Box,NONE)

global type is_Pause_Box(object x)
   return live_instance_check(x,Pause_Box)   
end type          

function Pause_Box_new_0()
   entity new
   new=call_method(super(),"new",NONE)
   VOID=call_method(new,"Set_Text",NONE)
   return new
end function

method("new",0,CLASS,NORMAL,PUBLIC,routine_id("Pause_Box_new_0"),NONE)

function Pause_Box_Set_Text_0()
   return call_method(super(),"Set_Text",{"Press any key to continue . . ."})
end function

method("Set_Text",0,INSTANCE,FINAL,PRIVATE,
routine_id("Pause_Box_Set_Text_0"),NONE)
method("Set_Text",1,INSTANCE,NORMAL,PUBLIC,NULL_METHOD,NONE)

function Pause_Box_Display_0()
   VOID=call_method(super(),"Display",NONE)
   while get_key()!=-1 do end while
   VOID=wait_key()
   return NIL
end function

method("Display",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Pause_Box_Display_0"),NONE)

end_class()  
          
          
global constant Input_Box=class("Input_Box",NORMAL,Text_Box,NONE) 

global type is_Input_Box(object x)
   return live_instance_check(x,Input_Box)
end type                             

property("Minimum",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,0)
property("Maximum",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,78)
property("Echo",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,0)

function Input_Box_new_0()
   entity new
   new=call_method(super(),"new",NONE)
   VOID=call_method(new,"Set_Text",NONE)
   return new
end function 

method("new",0,CLASS,NORMAL,PUBLIC,routine_id("Input_Box_new_0"),NONE) 

function Input_Box_Set_Text_0()
   return call_method(super(),"Set_Text",
   {repeat(HARD_SPACE,get_property(this(),"Maximum"))})
end function                                                             

method("Set_Text",0,INSTANCE,FINAL,PRIVATE,
routine_id("Input_Box_Set_Text_0"),NONE)
method("Set_Text",1,INSTANCE,NORMAL,PUBLIC,NULL_METHOD,NONE)

function Input_Box_Set_Limits_2(is_input_limit min,is_input_limit max)
   if not success() then return NIL end if
   if max<min then 
      throw(Type_Check_Failure)
      return NIL
   end if
   set_property(this(),"Minimum",min)
   set_property(this(),"Maximum",max)
   VOID=call_method(this(),"Set_Text",NONE)
   return NIL
end function

method("Set_Limits",2,INSTANCE,NORMAL,PUBLIC,
routine_id("Input_Box_Set_Limits_2"),NONE)

function Input_Box_Set_Echo_1(object echo)
   while TRUE do
      if not integer(echo) then exit end if
      if echo<32 then
         if echo!=0 then exit end if
      end if   
      if echo>255 then exit end if
      set_property(this(),"Echo",echo) 
      exit
   end while  
   return NIL
end function                          

method("Set_Echo",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Input_Box_Set_Echo_1"),{0})
method("Set_Echo",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Input_Box_Set_Echo_1"),NONE)

function Input_Box_Display_0()
   integer col,count,echo,key,line,min,max
   sequence pos,resp
   VOID=call_method(super(),"Display",NONE)
   min=get_property(this(),"Minimum")
   max=get_property(this(),"Maximum")
   echo=get_property(this(),"Echo")
   col=floor((78-max)/2)+2
   pos=get_position()              
   line=pos[1]-1
   position(line,col)
   resp=NONE
   count=0
   while get_key()!=-1 do end while
   while TRUE do
      key=wait_key()
      if key=13 then
         if count>=min then exit end if
      elsif key=8 then
         if count>0 then
            count-=1
            resp=resp[1..count]
            position(line,col+count)
            puts(1," ")
            position(line,col+count)   
         end if                     
      else
         if count<=max then
            count+=1
            resp&=key  
            if echo=0 then
               puts(1,key)
            else
               puts(1,echo)
            end if
         end if
      end if   
   end while
   position(pos[1],pos[2])
   return resp
end function

method("Display",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Input_Box_Display_0"),NONE)

end_class()


global constant Yes_No_Box=class("Yes_No_Box",NORMAL,Text_Box,NONE)

global type is_Yes_No_Box(object x)
   return live_instance_check(x,Yes_No_Box)
end type                             

function Yes_No_Box_Set_Text_0(object text)
   return call_method(super(),"Set_Text",{text&" [Y/N]"})
end function                  

method("Set_Text",1,INSTANCE,FINAL,PUBLIC,
routine_id("Yes_No_Box_Set_Text_0"),NONE)

function Yes_No_Box_Display_0()
   entity color
   integer index,key
   sequence pos,sel         
   VOID=call_method(super(),"Display",NONE) 
   while get_key()!=-1 do end while
   index=-1
   while index=-1 do
      key=upper(wait_key())
      index=find(key,"NY")-1
   end while 
   color=call_method(Screen,"Get_Color",NONE)                                         
   sel="Selected "&key
   VOID=call_method(Screen,"Print_Line",{sel,CENTER,color})
   delay(1)
   return index
end function

method("Display",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Yes_No_Box_Display_0"),NONE)

end_class()


global constant Menu_Error=exception("Menu_Error",Exception)
global constant Too_Many_Items=exception("Too_Many_Items",Menu_Error)
global constant Too_Few_Items=exception("Too_Few_Items",Menu_Error)
global constant Item_Not_Found=exception("Item_Not_Found",Menu_Error)

global constant Menu=class("Menu",NORMAL,Box,NONE)

global type is_Menu(object x)
   return live_instance_check(x,Menu)
end type                             

property("Items",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)
property("Item_Colors",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)

method("new",0,CLASS,NORMAL,PUBLIC,SUPER_METHOD,NONE)

method("clone",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE) 

method("delete",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE) 

function Menu_Set_Text_1(is_printable_string text)
   if not success() then return NIL end if
   if length(text)<=78 then   
      return call_method(super(),"Set_Text",{text}) 
   end if
   throw(Type_Check_Failure)
   return NIL
end function

method("Set_Text",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Set_Text_1"),NONE)

function Menu_Add_Item_2(is_printable_string text,is_Color color)
   sequence Items,Item_Colors
   if not success() then return NIL end if
   if length(text)>76 then 
      throw(Type_Check_Failure)
      return NIL
   end if
   if equal(color,DEFAULT_COLOR) then
      color=call_method(this(),"Get_Color",NONE)
   end if                               
   Items=get_property(this(),"Items")
   Item_Colors=get_property(this(),"Item_Colors")
   if length(Items)=10 then
      throw(Too_Many_Items)
   else                            
       set_property(this(),"Items",append(Items,text))
       set_property(this(),"Item_Colors",append(Item_Colors,color))
   end if
   return NIL 
end function

method("Add_Item",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Add_Item_2"),{DEFAULT_COLOR})
method("Add_Item",2,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Add_Item_2"),NONE)

function Menu_Edit_Item_3(is_integer index,is_printable_string text,
is_Color color)
   sequence Items,Item_Colors
   if not success() then return NIL end if
   Items=get_property(this(),"Items")                      
   if index<1 or index>length(Items) or length(text)>76 then 
      throw(Type_Check_Failure)
      return NIL
   end if
   if equal(color,DEFAULT_COLOR) then
      color=call_method(this(),"Get_Color",NONE)
   end if                               
   Item_Colors=get_property(this(),"Item_Colors")
   Items[index]=text
   Item_Colors[index]=color
   set_property(this(),"Items",Items)
   set_property(this(),"Item_Colors",Item_Colors)
   return NIL
end function

method("Edit_Item",2,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Edit_Item_3"),{DEFAULT_COLOR})
method("Edit_Item",3,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Edit_Item_3"),NONE)

function Menu_Get_Item_1(is_integer index)
   sequence Items,Item_Colors      
   if not success() then return NIL end if
   Items=get_property(this(),"Items")
   if index<1 or index>length(Items) then 
      throw(Type_Check_Failure)
      return NIL     
   end if 
   Item_Colors=get_property(this(),"Item_Colors")
   return {Items[index],Item_Colors[index]}   
end function

method("Get_Item",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Get_Item_1"),NONE)

function Menu_Remove_Item_1(object index)
   integer len
   sequence returnValue,Items,Item_Colors
   returnValue=call_method(this(),"Get_Item",index)
   if not success() then return NIL end if
   Items=get_property(this(),"Items")
   Item_Colors=get_property(this(),"Item_Colors")
   len=length(Items)
   set_property(this(),"Items",Items[1..index-1]&Items[index+1..len])
   set_property(this(),"Item_Colors",
   Item_Colors[1..index-1]&Item_Colors[index+1..len])
   return NIL
end function

method("Remove_Item",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Menu_Remove_Item_1"),NONE)

function Menu_Display_0()
   sequence Items,Item_Colors,formatText,spaces,pos,test,color,sel
   integer numItems,width,key
   VOID=call_method(Screen,"Clear",NONE)
   Items=get_property(this(),"Items")
   numItems=length(Items)
   if numItems<2 then
      throw(Too_Few_Items)
      return NIL
   end if
   Item_Colors=get_property(this(),"Item_Colors")
   Items=append(Items,call_method(this(),"Get_Text",NONE))
   Item_Colors=append(Item_Colors,call_method(this(),"Get_Color",NONE))
   width=0
   for i=1 to numItems+1 do
      formatText=Items[i]
      for j=length(formatText) to 1 by -1 do
         if formatText[j]!=' ' then
            formatText=formatText[1..j]
            exit
         end if
         if j=1 then formatText=NONE end if
      end for
      for j=1 to length(formatText) do
         if formatText[j]!=' ' then
            formatText=formatText[j..length(formatText)]
            exit
         end if
      end for
      Items[i]=formatText
      if (length(Items[i])+2*(i<=numItems))>width then
         width=length(Items[i])+2*(i<=numItems)
      end if
   end for
   position(12-numItems,1)
   VOID=call_method(Screen,"Print_Line",{201&repeat(205,width)&187,CENTER,
   Item_Colors[numItems+1]})
   spaces=floor(({width,width+1}-length(Items[numItems+1]))/2)
   VOID=call_method(Screen,"Print_Line",{186&repeat(' ',spaces[1])
   &Items[numItems+1]&repeat(' ',spaces[2])&186,CENTER,
   Item_Colors[numItems+1]})
   VOID=call_method(Screen,"Print_Line",{204&205&209&repeat(205,width-2)
   &185,CENTER,Item_Colors[numItems+1]})
   for i=1 to numItems do
      spaces=floor(({width-2,width-1}-length(Items[i]))/2)
      VOID=call_method(Screen,"Print_Line",
      {186&sprintf("%d",remainder(i,10))&179&repeat(' ',width-2)&186,CENTER,
      Item_Colors[numItems+1]})
      pos=get_position()
      VOID=call_method(Screen,"Print_At",{repeat(' ',spaces[1])&Items[i]
      &repeat(' ',spaces[2]),pos[1],floor(43-width/2),Item_Colors[i]})
      if i<numItems then
         VOID=call_method(Screen,"Print_Line",{199&196&197
         &repeat(196,width-2)&182,CENTER,Item_Colors[numItems+1]})
      end if
   end for
   VOID=call_method(Screen,"Print_Line",{200&205&207&repeat(205,width-2)
   &188,CENTER,Item_Colors[numItems+1]})
   test="1234567890"
   test=test[1..numItems]
   while get_key()!=-1 do end while
   key=0
   while key=0 do
      key=find(wait_key(),test)
   end while
   color=call_method(Screen,"Get_Color",NONE)
   sel="Selected #"&sprintf("%d",remainder(key,10))
   VOID=call_method(Screen,"Print_Line",{sel,CENTER,color})
   delay(1)
   return key
end function

method("Display",0,INSTANCE,NORMAL,PUBLIC,routine_id("Menu_Display_0"),NONE)

end_class()


package("Diamond Default")














