-- dcontain.e v4.0.0 by Mike Nelson
-- Diamond Standard Class Library
-- Containers       
   
without trace
without warning

include wildcard.e -- for upper()
include dexcept.e 

function validate_index(object index,integer len,integer plusOne)
   while TRUE do
      if not integer(index) then exit end if
      if index>len+plusOne then exit end if
      if index>0 then return index end if
      index=len+1+index  
      if index<1 then exit end if
      if index>len+plusOne then exit end if
      return index 
   end while
   throw(Type_Check_Failure)
   return NIL
end function 
-- For these classes index 0 = length+1 (for insertions)
-- index -1 = end of sequence etc.

type is_positive_integer(object x)
   if integer(x) then
      if x>0 then return TRUE end if
   end if
   throw(Type_Check_Failure)
   return TRUE
end type


package("Diamond Standard Class Library") 

global constant Container_Error=exception("Container_Error",Exception)

global constant Container_Empty=exception("Container_Empty",Container_Error)

global constant Container_Full=exception("Container_Full",Container_Error)
 
global constant Duplicate_Element=
exception("Duplicate_Element",Container_Error)  

global constant Missing_Element=exception("Missing_Element",Container_Error)


global constant Minimal_Container=interface("Minimal_Container",
NONE,{"Count#0","Get#0","Insert#1","Remove#0"},NONE) 

global type is_Minimal_Container(object x)
   return live_instance_check(x,Minimal_Container)   
end type


global constant Container=
interface("Container",Minimal_Container,{"Find#1","Get#1","Remove#1"},NONE)

global type is_Container(object x)
   return live_instance_check(x,Container)   
end type                              


global constant Basic_Container=
class("Basic_Container",ABSTRACT,Entity,{Clonable,Minimal_Container})

global type is_Basic_Container_instance(object x)
   return live_instance_check(x,Basic_Container)   
end type                                

property("Data",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)

method("new",0,CLASS,NORMAL,PUBLIC,SUPER_METHOD,NONE)

method("clone",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE)

method("delete",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE) 

function Basic_Container_Get_Data_0()
   return get_property(this(),"Data")   
end function                         

method("Get_Data",0,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Get_Data_0"),NONE)

function Basic_Container_Count_0()
   return length(get_property(this(),"Data"))
end function

method("Count",0,INSTANCE,FINAL,PUBLIC,
routine_id("Basic_Container_Count_0"),NONE)

function Basic_Container_Get_1(object index)
   integer len                                  
   sequence Data                                
   Data=get_property(this(),"Data")      
   len=length(Data)
   if len=0 then
      throw(Container_Empty)
      return NIL
   end if                     
   index=validate_index(index,len,FALSE) 
   if not success() then return NIL end if
   return Data[index]
end function

method("Get",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Basic_Container_Get_1"),{-1})
method("Get",1,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Get_1"),NONE)

function Basic_Container_Set_2(object newValue,object index)
   integer len
   sequence Data  
   Data=get_property(this(),"Data")
   len=length(Data) 
   if len=0 then
      throw(Container_Empty)
      return NIL
   end if                     
   index=validate_index(index,len,FALSE)
   if not success() then return NIL end if
   Data[index]=newValue
   set_property(this(),"Data",Data)
   return NIL
end function

method("Set",2,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Set_2"),NONE)

function Basic_Container_Insert_2(object newValue,object index)
   integer len
   sequence Data                                        
   Data=get_property(this(),"Data")
   len=length(Data)
   index=validate_index(index,len,TRUE)
   if not success() then return NIL end if
   set_property(this(),"Data",Data[1..index-1]&{newValue}&Data[index..len])
   return NIL
end function

method("Insert",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Basic_Container_Insert_2"),{0})
method("Insert",2,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Insert_2"),NONE)

function Basic_Container_Remove_1(object index)
   integer len
   sequence Data
   object returnValue                       
   Data=get_property(this(),"Data")
   len=length(Data) 
   if len=0 then
      throw(Container_Empty)
      return NIL
   end if                     
   index=validate_index(index,len,FALSE)
   if not success() then return NIL end if
   returnValue=Data[index]
   set_property(this(),"Data",Data[1..index-1]&Data[index+1..len])
   return returnValue
end function

method("Remove",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Basic_Container_Remove_1"),{-1})
method("Remove",1,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Remove_1"),NONE)

function Basic_Container_Find_1(object target)
   return find(target,get_property(this(),"Data"))
end function

method("Find",1,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Find_1"),NONE)

function Basic_Container_Find_Last_1(object target)
   sequence Data
   Data=get_property(this(),"Data")
   for i=length(Data) to 1 by -1 do
      if equal(Data[i],target) then
         return i
      end if
   end for
   return NIL
end function

method("Find_Last",1,INSTANCE,NORMAL,PROTECTED,
routine_id("Basic_Container_Find_Last_1"),NONE)

end_class() 


global constant Stack=class("Stack",NORMAL,Basic_Container,NONE)

global type is_Stack(object x)
   return live_instance_check(x,Stack)   
end type         

end_class()                                
                                           

global constant Queue=class("Queue",NORMAL,Basic_Container,NONE) 

global type is_Queue(object x)
   return live_instance_check(x,Queue)   
end type                        

function Queue_Get_0()
   return call_method(super(),"Get",{1})
end function

method("Get",0,INSTANCE,NORMAL,PUBLIC,routine_id("Queue_Get_0"),NONE)

function Queue_Remove_0()
   return call_method(super(),"Remove",{1})   
end function                               

method("Remove",0,INSTANCE,NORMAL,PUBLIC,routine_id("Queue_Remove_0"),NONE)

end_class() 
 

global constant Sorted_Sequence=
class("Sorted_Sequence",NORMAL,Basic_Container,Container)

global type is_Sorted_Sequence(object x)
   return live_instance_check(x,Sorted_Sequence) 
end type          
 
property("Ignore_Case",INSTANCE,PROTECTED,PRIVATE,PERSISTENT,FALSE)
property("Descending",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,FALSE)
property("Unique",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,FALSE)


function Sorted_Sequence_new_3(is_boolean ignoreCase,is_boolean descending,
is_boolean unique)
   entity new
   if not success() then return Null_Instance end if
   new=call_method(super(),"new",NONE)
   set_property(new,"Ignore_Case",ignoreCase)
   set_property(new,"Descending",descending)
   set_property(new,"Unique",unique)
   return new
end function

method("new",1,CLASS,NORMAL,PUBLIC,
routine_id("Sorted_Sequence_new_3"),{FALSE,FALSE})
method("new",2,CLASS,NORMAL,PUBLIC,
routine_id("Sorted_Sequence_new_3"),{FALSE})
method("new",3,CLASS,NORMAL,PUBLIC,
routine_id("Sorted_Sequence_new_3"),NONE)

method("Get",1,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE)    

method("Custom_Compare",0,CLASS,NORMAL,PRIVATE,NULL_METHOD,{-1})

function ss_search(object target,object searchType)
   integer direction,min,max,ignore,index,mid,test 
   sequence Data 
   object custom 
   custom=call_method(this_class(),"Custom_Compare",NONE)
   if not integer(custom) then custom=-1 end if
   Data=call_method(super(),"Get_Data",NONE)  
   ignore=get_property(this(),"Ignore_Case") 
   if get_property(this(),"Descending")=TRUE then
      direction=-1
   else
      direction=1
   end if
   min=1
   max=length(Data)
   index=0
   while min<=max do
      mid=floor((min+max)/2)
      if custom>=0 then
         test=call_func(custom,{target,Data[mid]}) 
      elsif ignore=TRUE then
         test=compare(upper(target),upper(Data[mid]))
      else
         test=compare(target,Data[mid])
      end if
      test*=direction
      if test=0 then
         index=mid
         test=searchType
      end if
      if test>0 then
         min=mid+1
      else
         max=mid-1
      end if
   end while
   if index=0 then return -min end if
   return index
end function                    

function Sorted_Sequence_Find_1(object target)
   return ss_search(target,-1)   
end function

method("Find",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Sorted_Sequence_Find_1"),NONE)  

function Sorted_Sequence_Find_Last_1(object target)
   return ss_search(target,1)
end function

method("Find_Last",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Sorted_Sequence_Find_Last_1"),{1})

function Sorted_Sequence_Insert_1(object newValue)
   integer index
   index=ss_search(newValue,1)
   if index>0 then
      if get_property(this(),"Unique")=TRUE then
         throw(Duplicate_Element)
         return NIL
      end if
      index=index+1
   else
      index=-index
   end if
   VOID=call_method(super(),"Insert",{newValue,index})
   return index
end function

method("Insert",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Sorted_Sequence_Insert_1"),NONE)

method("Remove",1,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE) 

method("Get_Data",0,INSTANCE,FINAL,PROTECTED,NULL_METHOD,{NONE}) 

end_class()


global constant Indexed_Sequence=
class("Indexed_Sequence",NORMAL,Sorted_Sequence,NONE)

global type is_Indexed_Sequence(object x)
   return live_instance_check(x,Indexed_Sequence)   
end type          

property("Index",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,1)
property("Records",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)
property("Include_Index",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,TRUE)

function Indexed_Sequence_new_5(object ignoreCase,object descending,
object unique,is_positive_integer index,is_boolean includeIndex)
   entity new                    
   if not success() then return Null_Instance end if
   new=call_method(super(),"new",{ignoreCase,descending,unique})
   if success() then 
      set_property(new,"Index",index)
      set_property(new,"Include_Index",includeIndex)
   end if
   return new
end function

method("new",4,CLASS,NORMAL,PUBLIC,
routine_id("Indexed_Sequence_new_5"),{TRUE})
method("new",5,CLASS,NORMAL,PUBLIC,
routine_id("Indexed_Sequence_new_5"),NONE)  

function Indexed_Sequence_Get_1(object index)
   integer indexElement
   sequence Records
   object key,record 
   key=call_method(super(),"Get",{index})
   if not success() then return NIL end if
   Records=get_property(this(),"Records")
   record=Records[index]
   if get_property(this(),"Include_Index")=TRUE then
      indexElement=get_property(this(),"Index")
      record=record[1..indexElement-1]&{key}
      &record[indexElement..length(record)]   
   end if 
   return record
end function  

method("Get",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Indexed_Sequence_Get_1"),NONE)

function Indexed_Sequence_Insert_1(is_sequence newValue)
   integer index,indexElement,len 
   sequence Records
   if not success() then return NIL end if
   indexElement=get_property(this(),"Index")  
   len=length(newValue)
   if len<indexElement then
      throw(Type_Check_Failure)
      return NIL
   end if
   index=call_method(super(),"Insert",{newValue[indexElement]})
   if not success() then return NIL end if
   newValue=newValue[1..indexElement-1]&newValue[indexElement+1..len]
   Records=get_property(this(),"Records")
   Records=Records[1..index-1]&{newValue}&Records[index..length(Records)]
   set_property(this(),"Records",Records)
   return index
end function

method("Insert",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Indexed_Sequence_Insert_1"),NONE)   

function Indexed_Sequence_Remove_1(object index)
   integer indexElement
   sequence Records
   object key,record 
   key=call_method(super(),"Remove",{index})
   if not success() then return NIL end if
   Records=get_property(this(),"Records")
   record=Records[index]
   Records=Records[1..index-1]&Records[index+1..length(Records)]
   set_property(this(),"Records",Records)
   if get_property(this(),"Include_Index")=TRUE then
      indexElement=get_property(this(),"Index")
      record=record[1..indexElement-1]&{key}
      &record[indexElement..length(record)]   
   end if 
   return record
end function  

method("Remove",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Indexed_Sequence_Remove_1"),{-1})
method("Remove",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Indexed_Sequence_Remove_1"),NONE)

end_class() 


global constant Priority_Container=
class("Priority_Container",ABSTRACT,Basic_Container,NONE) 

global type is_Priority_Container(object x)
   return live_instance_check(x,Priority_Container)   
end type

property("Priority",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,Null_Instance)
property("Descending",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,FALSE)

function Priority_Container_new_1(is_boolean descending)
   entity new,priority 
   if not success() then return Null_Instance end if
   priority=call_method(Sorted_Sequence,"new",{FALSE,descending,FALSE})
   new=call_method(super(),"new",NONE)
   set_property(new,"Priority",priority)
   set_property(new,"Descending",descending)
   return new
end function

method("new",0,CLASS,NORMAL,PUBLIC,
routine_id("Priority_Container_new_1"),{FALSE})
method("new",1,CLASS,NORMAL,PROTECTED,
routine_id("Priority_Container_new_1"),NONE)

function Priority_Container_clone_0()
   entity clone
   clone=call_method(super(),"clone",NONE)
   set_property(clone,"Priority",call_method(get_property(clone,"Priority"),
   "clone",NONE))
   return clone
end function

method("clone",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Priority_Container_clone_0"),NONE)

function Priority_Container_delete_0()
   VOID=call_method(get_property(this(),"Priority"),"delete",NONE)
   return call_method(super(),"delete",NONE)
end function

method("delete",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Priority_Container_delete_0"),NONE)

function Priority_Container_Insert_2(object newValue,is_integer priority)
   integer index
   entity Priority 
   if not success() then return NIL end if
   Priority=get_property(this(),"Priority")
   index=call_method(Priority,"Find",{priority})
   if index>0 then
      index+=1
   else
      index=-index
   end if
   VOID=call_method(Priority,"Insert",{priority})
   return call_method(super(),"Insert",{newValue,index})
end function

method("Insert",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Priority_Container_Insert_2"),{0})
method("Insert",2,INSTANCE,NORMAL,PUBLIC,
routine_id("Priority_Container_Insert_2"),NONE)

function Priority_Container_Remove_0()
   sequence superParams
   if get_property(this(),"Descending")=TRUE then
      superParams={1}
   else
      superParams=NONE
   end if
   VOID=call_method(get_property(this(),"Priority"),"Remove",superParams)
   return call_method(super(),"Remove",superParams)
end function

method("Remove",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Priority_Container_Remove_0"),NONE)

end_class() 


global constant Priority_Stack=
class("Priority_Stack",NORMAL,Priority_Container,NONE)

global type is_Priority_Stack(object x)
   return live_instance_check(x,Priority_Stack)   
end type          

end_class() 


global constant Priority_Queue=
class("Priority_Queue",NORMAL,Priority_Container,NONE)

global type is_Priority_Queue(object x)
   return live_instance_check(x,Priority_Queue)   
end type

function Priority_Queue_new_0()
   return call_method(super(),"new",{TRUE})
end function

method("new",0,CLASS,NORMAL,PUBLIC,routine_id("Priority_Queue_new_0"),NONE)

end_class()


package("Diamond Default")


