-- Diamond 4.0.0 by Mike Nelson

-- A Java-like class system for Euphoria with single inheritance and
-- interfaces.      
-- Exception handling for errors.  
-- Separate read and write access for properties.  
-- Polymorphic methods. 
-- Automatic method calls.  
-- Events with multiple handlers.    
-- Exception throwing type checks.
-- Automatic getters and setter for properties.
-- Extensive class information functions. 
-- Attempted method calls, property access, etc.
-- Inner classes.


without trace
without warning 

-- CONSTANTS:
-- Entity marker:
constant MARKER=-power(2,30)  
-- Events: 
global constant CURRENT_HANDLERS=-7
global constant DEFAULT_HANDLER=-6
global constant NULL_HANDLER=-5 
-- Methods:
global constant ABSTRACT_METHOD=-4
global constant NULL_METHOD=-3 
global constant SUPER_METHOD=-2
constant BAD_METHOD=-1
-- Meaningless return values:
global constant NIL=0
-- Booleans:
global constant FALSE=0
global constant TRUE=1  
-- Property and method varieties:
global constant INSTANCE=1
global constant CLASS=2
-- Class types:
global constant NORMAL=3      -- valid method type
global constant FINAL=4       -- valid method type
global constant ABSTRACT=5    -- valid method type    
global constant INTERFACE=6
global constant EXCEPTION=7
global constant NULL_CLASS=8 
-- Access control:
global constant PUBLIC=9
global constant PROTECTED=10
global constant PRIVATE=11
-- Program contexts:  
constant CLASS_DEFINITION=12 
constant MAIN_PROGRAM=13 
constant INSTANCE_EVENT=14
constant CLASS_EVENT=15
constant INSTANCE_METHOD=16
constant CLASS_METHOD=17                         
-- Serialization:
global constant PERSISTENT=18
global constant TRANSIENT=19  
-- Empty lists:
global constant NONE={}
-- Program stack:
constant STACK_EXPAND=repeat(NIL,16)
-- Identifiers:
constant ID_INIT="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
constant ID_DIGIT="0123456789"
constant ID_CHAR=ID_INIT&ID_DIGIT&'_'
-- Kernel entities:
global constant Entity={1,0,MARKER}
global constant Exception={2,0,MARKER}
global constant Null_Class={3,0,MARKER}
global constant Null_Instance={3,1,MARKER} 
constant Null_Class_ID=3
global constant Clonable={4,0,MARKER}
global constant Do_Not_Clone={5,0,MARKER}
constant Do_Not_Clone_ID=5
global constant Type_Check_Failure={6,0,MARKER}
constant Type_Check_Failure_ID=6   
-- Validation:
constant VALID_ENTITY_TYPE={INSTANCE,CLASS} 
constant VALID_METHOD_TYPE={NORMAL,FINAL,ABSTRACT}
constant VALID_CLASS_TYPE=VALID_METHOD_TYPE&{INTERFACE,EXCEPTION}
constant VALID_ACCESS={PUBLIC,PROTECTED,PRIVATE}
constant VALID_PERSISTENCE={PERSISTENT,TRANSIENT}
constant SPECIAL_HANDLERS={DEFAULT_HANDLER,NULL_HANDLER,CURRENT_HANDLERS} 
constant SUPERCLASS_MATRIX={NONE,NONE,{NORMAL,FINAL},NONE,
VALID_METHOD_TYPE,{INTERFACE},{EXCEPTION},NONE}
constant METHOD_TYPE_MATRIX={NONE,NONE,{NORMAL,FINAL},{FINAL},
VALID_METHOD_TYPE,{NORMAL,ABSTRACT},{ABSTRACT}}
constant OVERRIDE_MATRIX={NONE,NONE,{NORMAL,FINAL},NONE,VALID_METHOD_TYPE}
-- Error messages:
constant CONSTANT_NAME={"instance","class","normal","final","abstract",
"interface","exception","null","public","protected","private",
"class definition","main program","instance event","class event",
"instance method","class method","persistent","transient"}
constant CONSTANT_CAP_NAME={"Instance","Class","Normal","Final","Abstract",
"Interface","Exception","Null","Public","Protected","Private",
"Class definition","Main program","Instance event","Class event",
"Instance method","Class method","Persistent","Transient"} 
constant CONSTANT_UPPER_NAME={"INSTANCE","CLASS","NORMAL","FINAL","ABSTRACT",
"INTERFACE","EXCEPTION","NULL","PUBLIC","PROTECTED","PRIVATE",
"CLASS DEFINITION","MAIN_PROGRAM","INSTANCE EVENT","CLASS EVENT",
"INSTANCE METHOD","CLASS METHOD","PERSISTENT","TRANSIENT"}  
-- Debug file options:
global constant DFO_NONE=0
global constant DFO_WRITE=1
global constant DFO_APPEND=2
global constant DFO_NEW=3 
-- Serialization: 
global constant CURRENT_VERSION={4,0,0} 
constant VERSION_LIST={CURRENT_VERSION,{3,3,0},{3,2,0},{3,1,1},{3,1,0},
{3,0,1},{3,0,0},{2,2,0},{2,1,1},{2,1,0},{2,0,0},{1,2,0}} 
constant MEM=machine_func(16,4) -- allocate()
constant MAX_2B=power(2,15)
constant MAX_3B=power(2,23) 
constant SESSION_ID=date()&MEM&peek4u(MEM)  
-- Misc.: 
constant DOUBLE_NONE={NONE,NONE} 
constant TRIPLE_NONE=repeat(NONE,3)  
constant TRIPLE_DOUBLE_NONE=repeat(DOUBLE_NONE,3)  
constant ADD_NONE={NONE}
constant ADD_DOUBLE_NONE={DOUBLE_NONE}

-- VARIABLES:
-- Trash can for unwanted return values:
integer SETVOID_ID

type void_object(object x)
   if not equal(x,NIL) then call_proc(SETVOID_ID,{}) end if
   return TRUE
end type    

global void_object VOID

procedure setvoid()
   VOID=NIL
end procedure

SETVOID_ID=routine_id("setvoid")  

VOID=0  

-- Class database:
sequence ClassName ClassName={{"Entity",1},{"Exception",1},{"Null_Class",1}} 
sequence FullyDefined FullyDefined=repeat(TRUE,3)
sequence ClassType ClassType={ABSTRACT,EXCEPTION,NULL_CLASS}
sequence SuperclassList SuperclassList={{1},{2},NONE} 
sequence OuterClassList OuterClassList=TRIPLE_NONE 
sequence InnerClassList InnerClassList=TRIPLE_NONE 
sequence AccessList AccessList=SuperclassList
sequence InterfaceList InterfaceList=TRIPLE_NONE     
sequence PropertyName PropertyName=TRIPLE_DOUBLE_NONE
sequence PropertySource PropertySource=TRIPLE_DOUBLE_NONE
sequence PropertyReadAccess PropertyReadAccess=TRIPLE_DOUBLE_NONE  
sequence PropertyGetter PropertyGetter=TRIPLE_DOUBLE_NONE
sequence PropertyWriteAccess PropertyWriteAccess=TRIPLE_DOUBLE_NONE
sequence PropertySetter PropertySetter=TRIPLE_DOUBLE_NONE
sequence PropertyPersistence PropertyPersistence=TRIPLE_DOUBLE_NONE
sequence PropertyDefault PropertyDefault=TRIPLE_DOUBLE_NONE
sequence MethodName MethodName=TRIPLE_DOUBLE_NONE
sequence MethodSource MethodSource=TRIPLE_DOUBLE_NONE
sequence MethodType MethodType=TRIPLE_DOUBLE_NONE
sequence MethodAccess MethodAccess=TRIPLE_DOUBLE_NONE
sequence MethodID MethodID=TRIPLE_DOUBLE_NONE 
sequence MethodDefault MethodDefault=TRIPLE_DOUBLE_NONE
sequence EventName EventName=TRIPLE_DOUBLE_NONE
sequence EventSource EventSource=TRIPLE_DOUBLE_NONE
sequence EventLinkAccess EventLinkAccess=TRIPLE_DOUBLE_NONE
sequence EventLinker EventLinker=TRIPLE_DOUBLE_NONE 
sequence EventRaiseAccess EventRaiseAccess=TRIPLE_DOUBLE_NONE 
sequence EventRaiser EventRaiser=TRIPLE_DOUBLE_NONE
-- Entity database:
integer InstanceCounter InstanceCounter=1
sequence InstanceEntity InstanceEntity={NONE}&repeat(NIL,127) 
integer ClassCounter ClassCounter=0
sequence ClassEntity ClassEntity=TRIPLE_NONE
sequence PrototypeInstance PrototypeInstance=TRIPLE_NONE
-- Packages:
sequence PackageName PackageName=NONE
integer CurrentPackage CurrentPackage=NIL
-- Program stack:
integer ProgramCounter ProgramCounter=1
sequence Context Context=STACK_EXPAND Context[1]=MAIN_PROGRAM
sequence Caller Caller=STACK_EXPAND
sequence This This=STACK_EXPAND This[1]=Null_Instance
sequence ThisMethod ThisMethod=STACK_EXPAND ThisMethod[1]=NONE
sequence ThisEvent ThisEvent=STACK_EXPAND ThisEvent[1]=NONE
-- Exceptions:
integer PendingException PendingException=NIL
integer CaughtException CaughtException=NIL   
-- Events:
integer CancelEvent CancelEvent=FALSE
-- Serialization:
sequence ByteSequence ByteSequence=NONE
sequence PackageList PackageList=NONE
sequence ClassList ClassList=NONE 
sequence EntityList EntityList=NONE 
integer DSPointer DSPointer=1 
integer CloneOK CloneOK=TRUE 
integer BadClone BadClone=FALSE 
-- Error handling:
sequence ErrorText ErrorText=NONE
integer DEBUG_FILE
integer DebugFileOpt DebugFileOpt=DFO_WRITE             
integer DebugFromFatal DebugFromFatal=FALSE
integer DebugOpenFail DebugOpenFail=FALSE
integer ErrorScreenWidth ErrorScreenWidth=80 
-- Temporaries for class definition:
sequence TempMethods TempMethods=DOUBLE_NONE
sequence TempMethodsStack TempMethodsStack=NONE  
integer CurrentDef CurrentDef=NIL
-- Action flags:
integer TestAction TestAction=FALSE
integer TryAction TryAction=FALSE 
integer SystemAction SystemAction=FALSE   
object ActionReturn ActionReturn=NIL

-- ROUTINES:
-- Error handling:
function get_class_name(integer index)
   sequence theClass
   object thePackage
   theClass=ClassName[index] 
   thePackage=theClass[2]
   if integer(thePackage) then
      return theClass[1]&" ("&PackageName[thePackage]&")"    
   end if
   index=find(thePackage,ClassName)
   return theClass[1]&" in "&get_class_name(index)  
end function

function context_string()
   integer context,thisClass
   sequence cs 
   if PendingException!=NIL then
      cs="With "&get_class_name(PendingException)&" exception pending\ni"
   elsif CaughtException!=NIL then
      cs="With "&get_class_name(CaughtException)&" exception caught\ni"
   else
      cs="I"
   end if
   for i=ProgramCounter to 1 by -1 do
      if i=ProgramCounter then cs&="n " else cs&=",\ncalled from " end if
      context=Context[i]
      if context!=MAIN_PROGRAM then
         thisClass=Caller[i]
         cs&=get_class_name(thisClass)&" "
      end if
      cs&=CONSTANT_NAME[context]
      if context>=INSTANCE_METHOD then
         cs&=" "&ThisMethod[i] 
      elsif context>=INSTANCE_EVENT then
         cs&=" "&ThisEvent[i]
      end if
   end for
   return cs&"."
end function

type message_string(object x)
    object char
    if atom(x) then return FALSE end if
    for i=1 to length(x) do
        char=x[i]
        if not integer(char) then return FALSE end if
        if char<32 then return FALSE end if
        if char>255 then return FALSE end if
    end for
    return TRUE
end type  

type extended_message_string(object x)
   integer index
   if atom(x) then return FALSE end if
   while TRUE do
      index=find('\n',x)
      if index=NIL then exit end if
      if not message_string(x[1..index-1]) then return FALSE end if
      x=x[index+1..length(x)]
   end while
   return message_string(x)
end type  

function wrap_line(sequence line)
   integer offset,len
   len=length(line)  
   if len=0 then return "\n" end if
   offset=0
   while len>ErrorScreenWidth-1 do 
      for i=ErrorScreenWidth-1 to 1 by -1 do
         if line[i+offset]=' ' then
            line[i+offset]='\n'
            offset+=i
            len-=i
            exit
         end if
      end for
   end while 
   if line[length(line)]!='\n' then line&='\n' end if
   return line   
end function


global procedure error_screen_width(object width)
   if not integer(width) then return end if
   if width<20 then return end if
   ErrorScreenWidth=width
end procedure

global procedure fatal_error(object ErrorMessage)
   integer index 
   sequence errmess 
   if not extended_message_string(ErrorMessage) then
      ErrorMessage="*** Invalid error message. ***"
   end if
   ErrorText=ErrorMessage
   errmess="DIAMOND "&sprintf("%d.%d.%d",CURRENT_VERSION)
   &" FATAL ERROR\n"&ErrorMessage&"\n"&context_string() 
   ErrorMessage=NONE
   while length(errmess)!=0 do
      index=find('\n',errmess)
      if index=NIL then 
         ErrorMessage&=wrap_line(errmess)
         errmess=NONE
      else
         ErrorMessage&=wrap_line(errmess[1..index-1]) 
         errmess=errmess[index+1..length(errmess)]
      end if
   end while    
   if DebugOpenFail=FALSE then
      DebugFromFatal=TRUE
      call_proc(DEBUG_FILE,{})
      ErrorMessage&=ErrorText   
   end if
   machine_proc(37,ErrorMessage) -- crash_message
   ? 1/0 -- to generate ex.err
end procedure                        

global procedure debug_file_option(object option)
   if not integer(option) then return end if
   if option<DFO_NONE then return end if
   if option>DFO_NEW then return end if
   DebugFileOpt=option
end procedure

global procedure debug_file()
   integer appFlag,fn,classType,context,thisClass  
   sequence d,errTxt,filename,mode                     
   if DebugFileOpt=DFO_NONE then 
      ErrorText="No debug listing created."
      return 
   end if
   errTxt=ErrorText
   filename="diamond.err"                     
   if DebugFileOpt=DFO_WRITE then 
      mode="w"         
      appFlag=FALSE
   elsif DebugFileOpt=DFO_APPEND then
      mode="a"                      
      appFlag=sequence(machine_func(22,"diamond.err")) -- dir() 
   else
      mode="w" 
      appFlag=FALSE
      for i=1 to 999 do
         if atom(machine_func(22,filename)) then exit end if -- dir()
         filename="diamd"&sprintf("%03d",i)&".err"    
         if i=999 then DebugOpenFail=TRUE end if
      end for 
   end if
   if DebugOpenFail=TRUE then
      fn=-1
   else   
      fn=open(filename,mode)
   end if
   if fn=-1 then
      DebugOpenFail=TRUE
      if DebugFromFatal=FALSE then
         fatal_error("debug_file()\nCannot open "&filename
         &" debug file.")
      else
         ErrorText="*** Cannot open "&filename&" debug file. ***"
         return
      end if
   end if
   if appFlag=TRUE then
      ErrorText="Debug listing appended to "&filename&" debug file." 
   else
      ErrorText="New debug file "&filename&" created."
   end if
   d=date()
   printf(fn,"--- DIAMOND v%d.%d.%d DIAGNOSTIC LISTING FOR "
   &"%02d/%02d/%04d %02d:%02d:%02d ---",
   CURRENT_VERSION&{d[2],d[3],d[1]+1900,d[4],d[5],d[6]})
   if length(errTxt)!=0 then
      puts(fn,"\n\n\n--- FATAL ERROR ---\n\n")
      puts(fn,errTxt)
   end if
   if PendingException!=NIL then
      puts(fn,"\n\n\n--- PENDING EXCEPTION ---\n\n")
      puts(fn,get_class_name(PendingException))
   end if
   if CaughtException!=NIL then
      puts(fn,"\n\n\n--- CAUGHT EXCEPTION ---\n\n")
      puts(fn,get_class_name(CaughtException))
   end if
   puts(fn,"\n\n\n--- PROGRAM STACK ---\n\n")
   for i=ProgramCounter to 1 by -1 do
      printf(fn,"%2d  ",i)
      context=Context[i]
      if context!=MAIN_PROGRAM then
         thisClass=Caller[i]
         printf(fn,"%s ",{get_class_name(thisClass)})
      end if
      puts(fn,CONSTANT_NAME[context]) 
      if context>=INSTANCE_METHOD then
         puts(fn," "&ThisMethod[i])   
      elsif context>=INSTANCE_EVENT then
         puts(fn," "&ThisEvent[i])
      end if
      puts(fn,"\n")
   end for
   puts(fn,"\n\n--- CLASSES ---\n")
   for i=1 to ClassCounter do
      printf(fn,"\n\n--- %s ---\n\n",{get_class_name(i)})
      printf(fn,"CLASS ID:  %d\n",i) 
      classType=ClassType[i]
      if classType!=NIL then
         printf(fn,"CLASS TYPE:  %s\n",{CONSTANT_NAME[classType]})
      end if
      for j=2 to length(SuperclassList[i]) do
         if j=2 then
            puts(fn,"SUPERCLASSES:  ")
         else
            puts(fn,", ")
         end if
         puts(fn,get_class_name(SuperclassList[i][j]))
      end for
      if length(SuperclassList[i])>1 then puts(fn,"\n") end if
      for j=1 to length(InterfaceList[i]) do
         if j=1 then
            puts(fn,"INTERFACES:  ")
         else
            puts(fn,", ")
         end if
         puts(fn,get_class_name(InterfaceList[i][j]))
      end for
      if length(InterfaceList[i])>0 then puts(fn,"\n") end if 
      for j=INSTANCE to CLASS do
         for k=1 to length(PropertyName[i][j]) do 
            printf(fn,"%s PROPERTY:  %s  %s  %s",
            {CONSTANT_UPPER_NAME[j],
            PropertyName[i][j][k],
            get_class_name(PropertySource[i][j][k]),
            CONSTANT_NAME[PropertyReadAccess[i][j][k]]})
            if PropertyGetter[i][j][k]!=NULL_METHOD then
               puts(fn," getter")   
            end if              
            printf(fn,"  %s",
            {CONSTANT_NAME[PropertyWriteAccess[i][j][k]]})
            if PropertySetter[i][j][k]!=NULL_METHOD then
               puts(fn," setter")   
            end if
            printf(fn,"  %s\n",
            {CONSTANT_NAME[PropertyPersistence[i][j][k]]})
         end for
      end for           
      for j=INSTANCE to CLASS do
         for k=1 to length(MethodName[i][j]) do 
            printf(fn,"%s METHOD:  %s  %s  %s  %s\n",
            {CONSTANT_UPPER_NAME[j],
            MethodName[i][j][k],
            get_class_name(MethodSource[i][j][k]),
            CONSTANT_NAME[MethodType[i][j][k]],
            CONSTANT_NAME[MethodAccess[i][j][k]]})
         end for
      end for
      for j=INSTANCE to CLASS do
         for k=1 to length(EventName[i][j]) do
            printf(fn,"%s EVENT:  %s  %s  %s",
            {CONSTANT_UPPER_NAME[j],
            EventName[i][j][k],
            get_class_name(EventSource[i][j][k]),
            CONSTANT_NAME[EventLinkAccess[i][j][k]]})
            if EventLinker[i][j][k]!=NULL_METHOD then
               puts(fn," linker")   
            end if              
            printf(fn," %s",{CONSTANT_NAME[EventRaiseAccess[i][j][k]]})
            if EventRaiser[i][j][k]!=NULL_METHOD then
               puts(fn," raiser")   
            end if 
            puts(fn,"\n")  
         end for
      end for
   end for 
   printf(fn,"\n\n--- END OF LISTING FOR %02d/%02d/%04d %02d:%02d:%02d "
   &"--\n\n\n",{d[2],d[3],d[1]+1900,d[4],d[5],d[6]})
   close(fn)
end procedure

DEBUG_FILE=routine_id("debug_file")   

-- Program stack:
procedure expand_stack()
   Context&=STACK_EXPAND
   Caller&=STACK_EXPAND
   This&=STACK_EXPAND
   ThisMethod&=STACK_EXPAND
   ThisEvent&=STACK_EXPAND
end procedure

-- The identifier type:
global type identifier(object x)
   if atom(x) then return FALSE end if
   if length(x)=0 then return FALSE end if
   if not find(x[1],ID_INIT) then return FALSE end if
   for i=2 to length(x) do
      if not find(x[i],ID_CHAR) then return FALSE end if
   end for
   return TRUE
end type

type method_specifier(object x)
   object x2
   if atom(x) then return FALSE end if
   if length(x)!=2 then return FALSE end if
   if not identifier(x[1]) then return FALSE end if
   x2=x[2]
   if not integer(x2) then return FALSE end if
   return x2>=0
end type 

type method_name(object x)
   integer char,index,len,zflag
   if atom(x) then return FALSE end if
   index=find('#',x)
   if index=NIL then return FALSE end if
   len=length(x)
   if index=len then return FALSE end if
   if not identifier(x[1..index-1]) then return FALSE end if
   zflag=FALSE
   for i=index+1 to len do
      char=x[1]
      if not find(x[i],ID_DIGIT) then return FALSE end if   
      if x[i]='0' then
         if zflag=FALSE then 
            if i!=len then return FALSE end if
         end if
      else
         zflag=TRUE
      end if
   end for
   return TRUE
end type

-- The entity type:
global type entity(object x)
   object xi
   if atom(x) then return FALSE end if
   if length(x)!=3 then return FALSE end if
   if not equal(x[3],MARKER) then return FALSE end if
   xi=x[1]
   if not integer(xi) then return FALSE end if
   if xi<1 then return FALSE end if
   if xi>ClassCounter then return FALSE end if
   xi=x[2]
   if not integer(xi) then return FALSE end if
   if xi<0 then return FALSE end if
   return xi<=InstanceCounter
end type

global type instance_entity(object x)
   if not entity(x) then return FALSE end if
   return x[2]!=0
end type   

global type live_instance(object x)
   if not instance_entity(x) then return FALSE end if
   return sequence(InstanceEntity[x[2]])
end type   

global type deleted_instance(object x)
   if not instance_entity(x) then return FALSE end if
   return atom(InstanceEntity[x[2]])
end type   


global type class_entity(object x)
   if not entity(x) then return FALSE end if
   return x[2]=0
end type      

-- Property access:
type accessor_specifier(object x)
   if atom(x) then return find(x,VALID_ACCESS)!=NIL end if
   if length(x)!=2 then return FALSE end if
   if not find(x[1],VALID_ACCESS) then return FALSE end if
   x=x[2]
   if not integer(x) then return FALSE end if
   return x>=0
end type

-- Class information:
global function get_class(object theEntity)
   if not entity(theEntity) then return Null_Class end if
   theEntity[2]=0
   return theEntity
end function  

global function class_name(object theEntity)       
   if not entity(theEntity) then return NONE end if
   return get_class_name(theEntity[1])
end function  

global function class_type(object theEntity)
   if not entity(theEntity) then return NULL_CLASS end if
   return ClassType[theEntity[1]]  
end function

global function same_class(object entity1,object entity2)
   if not entity(entity1) then return FALSE end if
   if not entity(entity2) then return FALSE end if
   return entity1[1]=entity2[1]
end function

global function extends(object entity1,object entity2)
   integer class1
   if not entity(entity1) then return FALSE end if
   if not entity(entity2) then return FALSE end if
   class1=entity1[1]
   return find(entity2[1],SuperclassList[class1])!=0
end function

global function implements(object entity1,object entity2)
   integer class1
   if not entity(entity1) then return FALSE end if
   if not entity(entity2) then return FALSE end if
   class1=entity1[1]
   return find(entity2[1],InterfaceList[class1])!=0
end function 

type cast_entity(object x)
   if Context[ProgramCounter]<INSTANCE_METHOD then return FALSE end if
   if atom(x) then return FALSE end if
   if length(x)!=4 then return FALSE end if                   
   if not entity(x[1..3]) then return FALSE end if
   return find(x[4],SuperclassList[Caller[ProgramCounter]])=2 
end type                             

global function an_instance(object x)
   if entity(x) then
      if ClassType[x[1]]<ABSTRACT then
         x[2]=1   
         return x
      end if
   end if
   return Null_Instance
end function  

global function this_method() 
   return ThisMethod[ProgramCounter]
end function    

global function this_event()
   return ThisEvent[ProgramCounter]
end function    

global function last_method()
   if ProgramCounter<3 then return NONE end if
   return ThisMethod[ProgramCounter-1]
end function  

global function last_event()
   if ProgramCounter<3 then return NONE end if
   return ThisEvent[ProgramCounter-1]
end function 

global function program_stack()
   integer j
   sequence ps                     
   if Context[ProgramCounter]<INSTANCE_EVENT then return NONE end if
   ps=repeat(NIL,ProgramCounter-1)
   for i=ProgramCounter to 2 by -1 do 
      j=ProgramCounter-i+1
      ps[j]={This[i],ThisMethod[i],ThisEvent[i]}   
   end for 
   return ps
end function 

global function class_data(object target)
   integer classType,thisClass
   sequence data,entry,list,temp                          
   if not entity(target) then return NONE end if
   thisClass=target[1]
   classType=ClassType[thisClass]
   data={get_class_name(thisClass),classType}
   temp=SuperclassList[thisClass]          
   if length(temp)>0 then
      temp=temp[2..length(temp)]   
   end if           
   for i=1 to length(temp) do
      temp[i]={temp[i],0,MARKER}   
   end for
   data=append(data,temp)
   temp=InterfaceList[thisClass]
   for i=1 to length(temp) do
      temp[i]={temp[i],0,MARKER}   
   end for
   data=append(data,temp)
   for i=INSTANCE to CLASS do 
      list=NONE
      temp=PropertyName[thisClass][i]
      for j=1 to length(temp) do
         entry={temp[j],{PropertySource[thisClass][i][j],0,MARKER},
         PropertyReadAccess[thisClass][i][j],
         PropertyGetter[thisClass][i][j]!=NULL_METHOD,
         PropertyWriteAccess[thisClass][i][j],
         PropertySetter[thisClass][i][j]!=NULL_METHOD,
         PropertyPersistence[thisClass][i][j]}
         list=append(list,entry)    
      end for
      data=append(data,list)
   end for
   for i=INSTANCE to CLASS do  
      list=NONE
      temp=MethodName[thisClass][i]
      for j=1 to length(temp) do
         entry={temp[j],{MethodSource[thisClass][i][j],0,MARKER},
         MethodType[thisClass][i][j],MethodAccess[thisClass][i][j]}
         list=append(list,entry)   
      end for
      data=append(data,list)
   end for
   for i=INSTANCE to CLASS do 
      list=NONE 
      temp=EventName[thisClass][i]
      for j=1 to length(temp) do
         entry={temp[j],{EventSource[thisClass][i][j],0,MARKER},
         EventLinkAccess[thisClass][i][j], 
         EventLinker[thisClass][i][j]!=NULL_METHOD,
         EventRaiseAccess[thisClass][i][j],
         EventRaiser[thisClass][i][j]!=NULL_METHOD}   
         list=append(list,entry)
      end for 
      data=append(data,list)
   end for 
   return data
end function  

-- Class operations:
global function this()
   if Context[ProgramCounter]<INSTANCE_EVENT then
      fatal_error("this()\nNot allowed.")
   end if
   return This[ProgramCounter]
end function 

global function this_class()  
   sequence thisEntity 
   if Context[ProgramCounter]<INSTANCE_EVENT then
      fatal_error("this_class()\nNot allowed.")
   end if
   thisEntity=This[ProgramCounter]
   thisEntity[2]=0
   return thisEntity
end function

global function super()
   integer caller
   sequence thisEntity   
   if Context[ProgramCounter]<INSTANCE_METHOD then
      fatal_error("super()\nNot allowed.")
   end if             
   thisEntity=This[ProgramCounter]
   caller=Caller[ProgramCounter]
   return thisEntity&SuperclassList[caller][2]
end function  

global function super_class()
   integer caller
   sequence thisEntity   
   if Context[ProgramCounter]<INSTANCE_METHOD then
      fatal_error("super_class()\nNot allowed.")
   end if             
   thisEntity=This[ProgramCounter]
   thisEntity[2]=0
   caller=Caller[ProgramCounter]
   return thisEntity&SuperclassList[caller][2]
end function   
 
global function last() 
            
   if Context[ProgramCounter]<INSTANCE_EVENT then
      fatal_error("last()\nNot allowed.")
   end if
   if Context[ProgramCounter-1]<INSTANCE_EVENT then 
      return Null_Instance     
   end if
   return This[ProgramCounter-1]   
end function

global function last_class()          
   sequence lastEntity
   if Context[ProgramCounter]<INSTANCE_EVENT then
      fatal_error("last_class()\nNot allowed.")
   end if   
   if Context[ProgramCounter-1]<INSTANCE_EVENT then 
      return Null_Class     
   end if
   lastEntity=This[ProgramCounter-1]
   lastEntity[2]=0
   return lastEntity   
end function    

function action_error(sequence message)
   integer index
   sequence messpre
   if TestAction then
      ActionReturn=FALSE
   elsif TryAction then 
      if SystemAction then
         messpre="system"
      else
         messpre="try"   
      end if
      index=find('\n',message)
      for i=index-1 to 1 by -1 do
         if message[i]='_' then
            message=messpre&"_"&message[1..i-1]
            &message[index-2..length(message)] 
            exit  
         end if
      end for 
      if Context[ProgramCounter]=CLASS_DEFINITION and SystemAction=FALSE then  
         fatal_error(message)   
      end if
      ActionReturn={FALSE,message} 
   else
      fatal_error(message)   
   end if            
   TestAction=FALSE
   TryAction=FALSE
   SystemAction=FALSE
   return NIL
end function   

function access_allowed(integer access,integer target,integer source) 
   integer caller
   if SystemAction=TRUE then return TRUE end if     
   if access=PUBLIC then return TRUE end if
   if Context[ProgramCounter]<INSTANCE_METHOD then return FALSE end if
   caller=Caller[ProgramCounter]
   if find(caller,AccessList[source]) then 
      return TRUE
   end if
   if access=PRIVATE then return FALSE end if
   if find(caller,AccessList[target]) then 
      return TRUE
   end if
   return find(This[ProgramCounter][1],InnerClassList[target])!=NIL 
end function
                                      
global function get_property(object target,object prop)
   integer access,context,ID,index,saveTry,source
   integer targetType,thisClass,thisInstance
   sequence propName
   object entityValue
   context=Context[ProgramCounter] 
   if context=CLASS_DEFINITION then
      return action_error("get_property()\nNot allowed.")
   end if
   if not entity(target) then
      return action_error("get_property()\nTarget is not an entity.")
   end if
   thisInstance=target[2] 
   if thisInstance=1 then
      if not TestAction then target=Null_Instance end if
   end if
   thisClass=target[1]
   if thisInstance!=0 then
      entityValue=InstanceEntity[thisInstance]
      if atom(entityValue) then
         return action_error("get_property()\nTarget is a deleted instance.")
      end if 
      targetType=INSTANCE  
   else  
      if FullyDefined[thisClass]=FALSE then
         return action_error("get_property()\nTarget class "
         &get_class_name(thisClass)&" is not fully defined.")   
      end if
      if ClassType[thisClass]>ABSTRACT then
         return action_error("get_property()\n"
         &CONSTANT_CAP_NAME[ClassType[thisClass]]&" class "
         &get_class_name(thisClass)&" not allowed as target.")    
      end if
      entityValue=ClassEntity[thisClass]
      targetType=CLASS
   end if   
   propName=PropertyName[thisClass][targetType]
   index=find(prop,propName)
   if index=NIL then
      if identifier(prop) then
         return action_error("get_property()\nClass "
         &get_class_name(thisClass)&" does not define "
         &CONSTANT_NAME[targetType]&" property "&prop&".")
      end if
      return action_error("get_property()\nInvalid property name.")
   end if  
   if not TestAction then entityValue=entityValue[index] end if 
   source=PropertySource[thisClass][targetType][index]
   access=PropertyReadAccess[thisClass][targetType][index]              
   ID=PropertyGetter[thisClass][targetType][index]
   if access_allowed(access,thisClass,source)=FALSE then
      return action_error("get_property()\nAccess to "
      &get_class_name(thisClass)&" "&CONSTANT_NAME[targetType]
      &" property "&prop&" denied.")
   end if
   if TestAction then  
      TestAction=FALSE
      ActionReturn=TRUE
      return NIL     
   end if
   saveTry=TryAction 
   TryAction=FALSE
   if ID!=NULL_METHOD then  
      ProgramCounter+=1
      if ProgramCounter>length(Context) then expand_stack() end if
      Context[ProgramCounter]=targetType+INSTANCE_METHOD-INSTANCE
      Caller[ProgramCounter]=source
      This[ProgramCounter]=target 
      ThisMethod[ProgramCounter]=prop&":get"
      ThisEvent[ProgramCounter]=NONE 
      entityValue=call_func(ID,{entityValue})
      ProgramCounter-=1       
   end if 
   if saveTry then
      ActionReturn={TRUE,entityValue}
      return NIL
   end if  
   return entityValue
end function  

global function can_get(object target,object prop)
   TestAction=TRUE
   VOID=get_property(target,prop)
   return ActionReturn
end function

global function can_get_list(object target)
   integer targetType,thisClass
   sequence list,temp
   if not entity(target) then return NONE end if
   thisClass=target[1]
   targetType=(target[2]=0)+1
   list=NONE 
   temp=PropertyName[thisClass][targetType]
   for i=1 to length(temp) do
      if can_get(target,temp[i])=TRUE then
         list=append(list,temp[i])   
      end if
   end for
   return list
end function   

global function try_get(object target,object prop)
   TryAction=TRUE
   VOID=get_property(target,prop)
   return ActionReturn   
end function 

global procedure set_property(object target,object prop,object newValue)
   integer access,context,ID,index,saveTry,source
   integer targetType,thisClass,thisInstance
   sequence propName 
   object entityValue
   context=Context[ProgramCounter]
   if context=CLASS_DEFINITION then
      VOID=action_error("set_property()\nNot allowed.")
      return
   end if
   if not entity(target) then
      VOID=action_error("set_property()\nTarget is not an entity.")
      return
   end if
   thisInstance=target[2] 
   if thisInstance=1 then
      if not TestAction then target=Null_Instance end if
   end if
   thisClass=target[1]
   if thisInstance!=0 then 
      entityValue=InstanceEntity[thisInstance]
      if atom(entityValue) then
         VOID=action_error("set_property()\nTarget is a deleted instance.")
         return
      end if
      targetType=INSTANCE
   else 
      if FullyDefined[thisClass]=FALSE then
         VOID=action_error("set_property()\nTarget class "
         &get_class_name(thisClass)&" is not fully defined.")   
      end if
      if ClassType[thisClass]>ABSTRACT then
         VOID=action_error("set_property()\n"
         &CONSTANT_CAP_NAME[ClassType[thisClass]]&" class "
         &get_class_name(thisClass)&" not allowed as target.")    
         return
      end if             
      entityValue=ClassEntity[thisClass]
      targetType=CLASS
   end if 
   propName=PropertyName[thisClass][targetType]   
   index=find(prop,propName)
   if index=NIL then
      if identifier(prop) then
         VOID=action_error("set_property()\nClass "&get_class_name(thisClass)
         &" does not define "&CONSTANT_NAME[targetType]&" property "
         &prop&".")
         return
      end if
      VOID=action_error("set_property()\nInvalid property name.") 
      return
   end if   
   if not TestAction then entityValue=entityValue[index] end if
   source=PropertySource[thisClass][targetType][index]
   access=PropertyWriteAccess[thisClass][targetType][index] 
   ID=PropertySetter[thisClass][targetType][index] 
   if access_allowed(access,thisClass,source)=FALSE then
      VOID=action_error("set_property()\nAccess to "
      &get_class_name(thisClass)&" "&CONSTANT_NAME[targetType]
      &" property "&prop&" denied.")
      return
   end if
   if TestAction then 
      TestAction=FALSE
      ActionReturn=TRUE
      return
   end if
   saveTry=TryAction 
   TryAction=FALSE  
   if ID!=NULL_METHOD then
      ProgramCounter+=1
      if ProgramCounter>length(Context) then expand_stack() end if
      Context[ProgramCounter]=targetType+INSTANCE_METHOD-INSTANCE
      Caller[ProgramCounter]=source
      This[ProgramCounter]=target 
      ThisMethod[ProgramCounter]=prop&":set"
      ThisEvent[ProgramCounter]=NONE 
      newValue=call_func(ID,{entityValue,newValue})
      ProgramCounter-=1       
   end if 
   if targetType=INSTANCE then
      InstanceEntity[thisInstance][index]=newValue
   else
      ClassEntity[thisClass][index]=newValue
   end if  
   if saveTry then
      ActionReturn={TRUE,NIL}
   end if  
end procedure    

global function can_set(object target,object prop)
   TestAction=TRUE
   set_property(target,prop,NIL)
   return ActionReturn   
end function 

global function can_set_list(object target)
   integer targetType,thisClass
   sequence list,temp
   if not entity(target) then return NONE end if
   thisClass=target[1]  
   targetType=(target[2]=0)+1 
   list=NONE         
   temp=PropertyName[thisClass][targetType]
   for i=1 to length(temp) do
      if can_set(target,temp[i])=TRUE then
         list=append(list,temp[i])   
      end if
   end for
   return list
end function

global function try_set(object target,object prop,object newValue)
   TryAction=TRUE
   set_property(target,prop,newValue)
   return ActionReturn   
end function 

global function call_method(object target,object meth,object Param)
   integer  access,context,ID,index,searchClass
   integer saveTry,source,targetType,thisClass,thisInstance
   sequence default,methName
   object methReturn 
   CaughtException=NIL 
   context=Context[ProgramCounter]
   if (context=CLASS_DEFINITION and SystemAction=FALSE) 
   or PendingException!=NIL then
      return action_error("call_method()\nNot allowed.")
   end if
   if entity(target) then 
      target&=target[1]
   elsif not cast_entity(target) then
      return action_error("call_method()\nTarget is not an entity.")
   end if 
   if atom(Param) then
      return action_error("call_method()\nParameter list must be a sequence.")
   end if                           
   thisInstance=target[2]
   if thisInstance=1 and not TestAction then 
      target[1]=Null_Class_ID 
      searchClass=Null_Class_ID
   else
      searchClass=target[4]
   end if                  
   thisClass=target[1]
   target=target[1..3]   
   if thisInstance!=0 then
      if atom(InstanceEntity[thisInstance]) then
         return action_error("call_method()\nTarget is a deleted instance.")
      end if
      targetType=INSTANCE
   else  
      if FullyDefined[thisClass]=FALSE then
         return action_error("call_method()\nTarget class "
         &get_class_name(thisClass)&" is not fully defined.")   
      end if   
      if ClassType[searchClass]>ABSTRACT then
         return action_error("call_method()\n"
         &CONSTANT_CAP_NAME[ClassType[searchClass]]&" class "
         &get_class_name(thisClass)&" not allowed as target.")    
      end if
      targetType=CLASS
   end if  
   methName=MethodName[searchClass][targetType]
   meth&=sprintf("#%d",length(Param))
   index=find(meth,methName)
   if index=NIL then
      if method_name(meth) then
         return action_error("call_method()\nClass "&get_class_name(searchClass)
         &" does not define "&CONSTANT_NAME[targetType]&" method "
         &meth&".")
      end if
      return action_error("call_method()\nInvalid method name.")
   end if 
   source=MethodSource[searchClass][targetType][index]
   access=MethodAccess[searchClass][targetType][index]
   ID=MethodID[searchClass][targetType][index]
   default=MethodDefault[searchClass][targetType][index]
   if ID=ABSTRACT_METHOD then
      return action_error("call_method()\nAccess to "
      &get_class_name(searchClass)&" abstract "&CONSTANT_NAME[targetType]
      &" method "&meth&" denied.")
   end if  
   if access_allowed(access,thisClass,source)=FALSE then
      return action_error("call_method()\nAccess to "
      &get_class_name(searchClass)&" "&CONSTANT_NAME[targetType]
      &" method "&meth&" denied.")
   end if    
   SystemAction=FALSE 
   if TestAction then 
      TestAction=FALSE
      ActionReturn=TRUE
      return NIL
   end if
   saveTry=TryAction 
   TryAction=FALSE   
   if ID=NULL_METHOD then 
      methReturn=default[1] 
   else
      ProgramCounter+=1
      if ProgramCounter>length(Context) then expand_stack() end if
      Context[ProgramCounter]=targetType+INSTANCE_METHOD-INSTANCE
      Caller[ProgramCounter]=source
      This[ProgramCounter]=target
      ThisMethod[ProgramCounter]=meth 
      ThisEvent[ProgramCounter]=NONE
      methReturn=call_func(ID,Param&default)
      ProgramCounter-=1  
   end if 
   if saveTry then
      ActionReturn={TRUE,methReturn}
      return NIL
   end if
   return methReturn
end function  

global function can_call(object target,object meth,object params)
   TestAction=TRUE
   VOID=call_method(target,meth,params)
   return ActionReturn
end function   

global function can_call_list(object target)
   integer index,targetType,thisClass
   sequence item,list,temp 
   object params
   if not entity(target) then return NONE end if
   thisClass=target[1] 
   targetType=(target[2]=0)+1
   list=NONE         
   temp=MethodName[thisClass][targetType]
   for i=1 to length(temp) do 
      item=temp[i]
      index=find('#',item)
      params=0
      for j=index+1 to length(item) do
         params=params*10+item[j]-'0'   
      end for  
      params=repeat(NIL,params)
      if can_call(target,item[1..index-1],params)=TRUE then
         list=append(list,item)   
      end if
   end for
   return list
end function        

global function try_call(object target,object prop,object params)
   TryAction=TRUE
   VOID=call_method(target,prop,params)
   return ActionReturn   
end function 

function system_call(object target,object prop,object params)
   TryAction=TRUE
   SystemAction=TRUE
   VOID=call_method(target,prop,params)   
   return ActionReturn
end function

global procedure raise_event(object target,object evt,object Param)
   integer access,context,idLen,index,RID,saveTry,source
   integer saveCE,targetType,thisClass,thisID,thisInstance
   sequence entityValue,evtName,evtsig  
   object cancel,ID                  
   CaughtException=NIL
   context=Context[ProgramCounter]
   if context=CLASS_DEFINITION or PendingException!=NIL then
      VOID=action_error("raise_event()\nNot allowed.") 
      return
   end if
   if not entity(target) then
      VOID=action_error("raise_event()\nTarget is not an entity.") 
      return
   end if 
   if atom(Param) then
      VOID=action_error("raise_event()\nParameter list must be a sequence.")
      return
   end if
   thisInstance=target[2] 
   if thisInstance=1 then
      if not TestAction then target=Null_Instance end if
   end if
   thisClass=target[1] 
   if thisInstance!=0 then 
      entityValue=InstanceEntity[thisInstance]
      if atom(entityValue) then
         VOID=action_error("raise_event()\nTarget is a deleted instance.") 
         return
      end if
      targetType=INSTANCE
   else   
      if FullyDefined[thisClass]=FALSE then
         VOID=action_error("raise_event()\nTarget class "
         &get_class_name(thisClass)&" is not fully defined.")   
      end if
      if ClassType[thisClass]>ABSTRACT then
         VOID=action_error("raise_event()\n"
         &CONSTANT_CAP_NAME[ClassType[thisClass]]&" class "
         &get_class_name(thisClass)&" not allowed as target.")
         return    
      end if
      entityValue=ClassEntity[thisClass]
      targetType=CLASS     
   end if
   evtName=EventName[thisClass][targetType] 
   evtsig=evt&sprintf("#%d",length(Param))
   index=find(evtsig,evtName)
   if index=NIL then
      if identifier(evt) then
         VOID=action_error("raise_event()\nClass "&get_class_name(thisClass)
         &" does not define "&CONSTANT_NAME[targetType]&" event "&evtsig&".")
         return
      end if
      VOID=action_error("raise_event()\nInvalid event name.") 
      return
   end if
   source=EventSource[thisClass][targetType][index]
   access=EventRaiseAccess[thisClass][targetType][index]   
   RID=EventRaiser[thisClass][targetType][index]
   if access_allowed(access,thisClass,source)=FALSE then
      VOID=action_error("raise_event()\nAccess to "
      &get_class_name(thisClass)&" "&CONSTANT_NAME[targetType]
      &" event "&evt&" denied.")
      return                          
   end if 
   if TestAction then
      TestAction=FALSE 
      ActionReturn=TRUE
      return 
   end if 
   saveTry=TryAction 
   TryAction=FALSE  
   saveCE=CancelEvent 
   if RID=NULL_METHOD then  
      cancel=FALSE
   else
      ProgramCounter+=1
      if ProgramCounter>length(Context) then expand_stack() end if
      Context[ProgramCounter]=targetType+INSTANCE_METHOD-INSTANCE
      Caller[ProgramCounter]=source
      This[ProgramCounter]=target 
      ThisMethod[ProgramCounter]=evtsig&":raise"
      ThisEvent[ProgramCounter]=NONE 
      cancel=call_func(RID,{Param})
      ProgramCounter-=1  
   end if 
   if not equal(cancel,FALSE) then 
      CancelEvent=saveCE
      if saveTry then
         ActionReturn={TRUE,NIL}   
      end if
      return
   end if                              
   CancelEvent=FALSE
   ID=entityValue[index+length(PropertyName[thisClass][targetType])] 
   idLen=length(ID)
   ProgramCounter+=1
   if ProgramCounter>length(Context) then expand_stack() end if
   Context[ProgramCounter]=targetType+INSTANCE_EVENT-INSTANCE
   Caller[ProgramCounter]=source
   This[ProgramCounter]=target 
   ThisMethod[ProgramCounter]=NONE
   ThisEvent[ProgramCounter]=evtsig  
   VOID=system_call(target,"Before_"&evt,Param)
   for i=1 to idLen+1 do
      if CancelEvent=TRUE then exit end if
      if PendingException!=NIL then exit end if
      if targetType=INSTANCE then 
         if atom(InstanceEntity[thisInstance]) then exit end if
      end if
      if i<=idLen then
         thisID=ID[i]                         
         if thisID=DEFAULT_HANDLER then
            VOID=system_call(target,"On_"&evt,Param)
         else
            call_proc(thisID,Param)        
         end if
      else  
         VOID=system_call(target,"After_"&evt,Param)
      end if   
   end for 
   CancelEvent=saveCE
   ProgramCounter-=1  
   if saveTry then
      ActionReturn={TRUE,NIL}   
   end if
end procedure   

global function can_raise(object target,object evt,object params)
   TestAction=TRUE
   raise_event(target,evt,params)
   return ActionReturn
end function  

global function can_raise_list(object target)
   integer index,targetType,thisClass
   sequence item,list,temp
   object params
   if not entity(target) then return NONE end if
   thisClass=target[1] 
   targetType=(target[2]=0)+1
   list=NONE         
   temp=EventName[thisClass][targetType]
   for i=1 to length(temp) do   
      item=temp[i]
      index=find('#',item)
      params=0
      for j=index+1 to length(item) do
         params=params*10+item[j]-'0'   
      end for 
      params=repeat(NIL,params)
      if can_raise(target,item[1..index-1],params)=TRUE then
         list=append(list,item)   
      end if
   end for
   return list
end function        

global function try_raise(object target,object evt,object params)
   TryAction=TRUE
   raise_event(target,evt,params)
   return ActionReturn   
end function 

global procedure link_handler(object target,object evt,object ID)
   integer access,context,index,index2,len,len2,LID,ptr
   integer saveTry,source,targetType,thisClass,thisInstance
   sequence evtName,ID2 
   object cancel,thisID
   context=Context[ProgramCounter]
   if context=CLASS_DEFINITION then
      VOID=action_error("link_handler()\nNot allowed.")
      return
   end if
   if not entity(target) then
      VOID=action_error("link_handler()\nTarget is not an entity.")
      return
   end if 
   thisInstance=target[2] 
   if thisInstance=1 then
      if not TestAction then target=Null_Instance end if
   end if
   thisClass=target[1]
   if thisInstance!=0 then
      if atom(InstanceEntity[thisInstance]) then
         VOID=action_error("link_handler()\nTarget is a deleted instance.")
         return
      end if  
      targetType=INSTANCE
   else  
      if FullyDefined[thisClass]=FALSE then
         VOID=action_error("link_handler()\nTarget class "
         &get_class_name(thisClass)&" is not fully defined.")   
      end if 
      if ClassType[thisClass]>ABSTRACT then
         VOID=action_error("link_handler()\n"
         &CONSTANT_CAP_NAME[ClassType[thisClass]]&" class "
         &get_class_name(thisClass)&" not allowed as target.")    
         return
      end if
      targetType=CLASS  
   end if 
   evtName=EventName[thisClass][targetType]
   if method_specifier(evt) then
      evt=evt[1]&sprintf("#%d",evt[2])      
   end if
   index=find(evt,evtName)
   if index=NIL then
      if method_name(evt) then
         VOID=action_error("link_handler()\nClass "&get_class_name(thisClass)
         &" does not define "&CONSTANT_NAME[targetType]&" event "&evt&".")
         return
      end if
      VOID=action_error("link_handler()\nInvalid event name.")
      return
   end if  
   index2=index+length(PropertyName[thisClass][targetType])      
   if atom(ID) then ID={ID} end if 
   ptr=1
   len=length(ID)
   while ptr<=len do
      for i=ptr to len do 
         thisID=ID[i]
         if not integer(thisID) 
         or (thisID<0 and not find(thisID,SPECIAL_HANDLERS)) then
            VOID=action_error("link_handler()\nHandler has invalid "
            &"routine id.")
            return
         end if             
         if thisID=NULL_HANDLER then
            ID=ID[1..i-1]&ID[i+1..len]
            ptr=i     
            len=-1
            exit
         elsif thisID=CURRENT_HANDLERS then
            if targetType=INSTANCE then
               ID2=InstanceEntity[thisInstance][index2] 
            else  
               ID2=ClassEntity[thisClass][index2]  
            end if
            ID=ID[1..i-1]&ID2&ID[i+1..len]
            len2=length(ID2)
            ptr=i+len2
            len+=len2-1
            exit
         end if 
         if i=len then ptr=i+1 end if
      end for 
   end while
   source=EventSource[thisClass][targetType][index]
   access=EventLinkAccess[thisClass][targetType][index]  
   if access_allowed(access,thisClass,source)=FALSE then
      VOID=action_error("link_handler()\nAccess to "
      &get_class_name(thisClass)&" "&CONSTANT_NAME[targetType]
      &" event "&evt&" denied.")
      return
   end if  
   if TestAction then   
      TestAction=FALSE
      ActionReturn=TRUE
      return
   end if            
   saveTry=TryAction
   TryAction=FALSE  
   LID=EventLinker[thisClass][targetType][index]
   if LID=NULL_METHOD then
      cancel=FALSE
   else
      ProgramCounter+=1
      if ProgramCounter>length(Context) then expand_stack() end if
      Context[ProgramCounter]=targetType+INSTANCE_METHOD-INSTANCE
      Caller[ProgramCounter]=source
      This[ProgramCounter]=target 
      ThisMethod[ProgramCounter]=evt&":link"
      ThisEvent[ProgramCounter]=NONE 
      cancel=call_func(LID,{ID})
      ProgramCounter-=1                  
   end if   
   if equal(cancel,FALSE) then
      if targetType=INSTANCE then 
         InstanceEntity[thisInstance][index2]=ID   
      else  
         ClassEntity[thisClass][index2]=ID
      end if    
   end if
   if saveTry then
      ActionReturn={TRUE,NIL}     
   end if
end procedure   

global function can_link(object target,object evt) 
   TestAction=TRUE 
   link_handler(target,evt,DEFAULT_HANDLER)
   return ActionReturn
end function    

global function can_link_list(object target)
   integer targetType,thisClass
   sequence list,temp
   if not entity(target) then return NONE end if
   thisClass=target[1]
   targetType=(target[2]=0)+1
   list=NONE         
   temp=EventName[thisClass][targetType]
   for i=1 to length(temp) do
      if can_link(target,temp[i])=TRUE then
         list=append(list,temp[i])   
      end if
   end for
   return list
end function   

global function try_link(object target,object evt,object params)
   TryAction=TRUE
   link_handler(target,evt,params)
   return ActionReturn   
end function 
     
global procedure cancel_event()
   if Context[ProgramCounter]<INSTANCE_EVENT then
      fatal_error("cancel_event()\nNot allowed.")   
   end if                                            
   CancelEvent=TRUE
end procedure


-- Class definition:  
global procedure package(object name) 
   integer index,len  
   if Context[ProgramCounter]!=MAIN_PROGRAM or PendingException!=NIL then
      fatal_error("package()\nNot allowed.")
   end if
   if not message_string(name) then
      fatal_error("package()\nInvalid package name.") 
   end if    
   len=length(name)
   for i=1 to len do
      if name[i]!=' ' then
         name=name[i..len]
         len-=i-1
         exit
      end if
      if i=len then
         name="Diamond Default"
         len=0
      end if
   end for
   if len>0 then
      for i=len to 1 by -1 do
         if name[i]!=' ' then
            name=name[1..i]
            len=i
            exit   
         end if
      end for      
      while TRUE do
         index=match("  ",name)
         if index=NIL then exit end if
         name=name[1..index]&name[index+2..length(name)]
      end while
   end if        
   index=find(name,PackageName)
   if index=NIL then
      PackageName&={name}
      CurrentPackage=length(PackageName)
   else                                 
      CurrentPackage=index
   end if
end procedure     

procedure new_class(sequence Name)
   integer index
   index=find(Name,ClassName)
   if index!=NIL then
      fatal_error("class()\n"&get_class_name(index)
      &" already defined.")   
   end if 
   ClassCounter+=1
   CurrentDef=ClassCounter 
   TempMethodsStack={TempMethods}&TempMethodsStack
   TempMethods=DOUBLE_NONE
   ClassName&={Name}  
   FullyDefined&=FALSE
   ClassType&=NIL
   SuperclassList&=ADD_NONE 
   OuterClassList&=ADD_NONE
   InnerClassList&=ADD_NONE 
   AccessList&=ADD_NONE
   InterfaceList&=ADD_NONE
   PropertyName&=ADD_DOUBLE_NONE
   PropertySource&=ADD_DOUBLE_NONE
   PropertyReadAccess&=ADD_DOUBLE_NONE
   PropertyGetter&=ADD_DOUBLE_NONE
   PropertyWriteAccess&=ADD_DOUBLE_NONE
   PropertySetter&=ADD_DOUBLE_NONE
   PropertyPersistence&=ADD_DOUBLE_NONE
   PropertyDefault&=ADD_DOUBLE_NONE
   MethodName&=ADD_DOUBLE_NONE
   MethodSource&=ADD_DOUBLE_NONE
   MethodType&=ADD_DOUBLE_NONE
   MethodAccess&=ADD_DOUBLE_NONE
   MethodID&=ADD_DOUBLE_NONE
   MethodDefault&=ADD_DOUBLE_NONE
   EventName&=ADD_DOUBLE_NONE
   EventSource&=ADD_DOUBLE_NONE
   EventLinkAccess&=ADD_DOUBLE_NONE 
   EventLinker&=ADD_DOUBLE_NONE
   EventRaiseAccess&=ADD_DOUBLE_NONE
   EventRaiser&=ADD_DOUBLE_NONE
   PrototypeInstance&=ADD_NONE
   ClassEntity&=ADD_NONE  
   ProgramCounter+=1
   if ProgramCounter>length(Context) then expand_stack() end if
   Context[ProgramCounter]=CLASS_DEFINITION
   Caller[ProgramCounter]=CurrentDef
   This[ProgramCounter]={CurrentDef,0,MARKER}
end procedure  


procedure merge_interface()
   integer index,superClass
   sequence meth,oldMeths,newMeths,supers
   supers=SuperclassList[ClassCounter]
   for i=1 to length(supers) do
      for j=INSTANCE to CLASS do
         oldMeths=MethodName[ClassCounter][j]
         superClass=supers[i]
         newMeths=MethodName[superClass][j]
         for k=1 to length(newMeths) do
            meth=newMeths[k]
            index=find(meth,oldMeths)
            if index=NIL then
               MethodName[ClassCounter][j]&={meth}
               MethodSource[ClassCounter][j]&=MethodSource[superClass][j][k]
               MethodType[ClassCounter][j]&=ABSTRACT
               MethodAccess[ClassCounter][j]&=PUBLIC                     
               MethodID[ClassCounter][j]&=ABSTRACT_METHOD
               MethodDefault[ClassCounter][j]&=ADD_NONE
            end if
         end for
      end for
   end for   
end procedure       

procedure subset_check(integer superClass,integer thisClass)
   sequence superList,thisList
   superList=OuterClassList[superClass]
   thisList=OuterClassList[thisClass] 
   if length(superList)>0 then 
      if match(superList,thisList)=NIL then
         fatal_error("class()\nSuperclass "&get_class_name(superClass)
         &" is an inner class with incompatible outer classes.")  
      end if
   end if
end procedure

global function class(object Name,object Type,object Extends,
object Implements)
   integer ilen,innerFlag,outerClass,outerType,superClass,slen
   integer thisClass
   object temp   
   CaughtException=NIL
   if Context[ProgramCounter]>MAIN_PROGRAM or PendingException!=NIL then
      fatal_error("class()\nNot allowed.")
   end if
   if not identifier(Name) then
      fatal_error("class()\nInvalid class name.")
   end if
   if CurrentDef=NIL then  
      innerFlag=FALSE
      new_class({Name,CurrentPackage}) 
   else       
      innerFlag=TRUE
      new_class({Name,ClassName[CurrentDef]})   
   end if  
   if not find(Type,VALID_CLASS_TYPE) then
      fatal_error("class()\nInvalid class type.")
   end if
   if innerFlag=TRUE then
      if Type>ABSTRACT then
         fatal_error("class()\n"&CONSTANT_CAP_NAME[Type]
         &" not allowed as inner class.")
      else  
         outerClass=Caller[ProgramCounter-1]
         outerType=ClassType[outerClass]
         if outerType>ABSTRACT then
            fatal_error("class()\nInner class not allowed in "
            &CONSTANT_NAME[outerType]&" class definition.")   
         end if                                         
      end if
   end if
   ClassType[CurrentDef]=Type 
   if atom(Extends) or entity(Extends) then
      Extends={Extends}   
   end if 
   slen=length(Extends) 
   if slen=0 then
      if Type<INTERFACE then
         Extends={Entity}
         slen=1
      elsif Type=EXCEPTION then
         Extends={Exception}
         slen=1
      end if
   end if  
   if Type!=INTERFACE then
      if slen>1 then
         fatal_error("class()\nMultiple superclasses not allowed for "
         &CONSTANT_NAME[Type]&" class.")
      end if
   end if
   for i=1 to slen do 
      temp=Extends[i]
      if not class_entity(temp) then
         fatal_error("class()\nInvalid superclass list.")     
      end if 
      temp=temp[1]
      Extends[i]=temp
      if find(temp,Extends[1..i-1]) then
         fatal_error("class()\nDuplicate superclasses.")       
      end if
   end for 
   for i=1 to slen do
      superClass=Extends[i] 
      if not find(Type,SUPERCLASS_MATRIX[ClassType[superClass]]) then
         fatal_error("class()\n"&CONSTANT_CAP_NAME[ClassType[superClass]]
         &" class "&get_class_name(superClass)&" may not be a "
         &"superclass of "&CONSTANT_NAME[Type]&" class.")
      end if 
   end for   
   if atom(Implements) or entity(Implements) then
      Implements={Implements}
   end if              
   ilen=length(Implements)
   if Type>ABSTRACT then
      if ilen!=0 then
         fatal_error("class()\n"&CONSTANT_CAP_NAME[Type]&" class "
         &get_class_name(ClassCounter)&" must have an empty interface list.")   
      end if
   end if   
   for i=1 to ilen do
      temp=Implements[i]
      if not class_entity(temp) then
         fatal_error("class()\nInvalid interface list.")     
      end if 
      temp=temp[1]
      Implements[i]=temp
      if find(temp,Implements[1..i-1]) then
         fatal_error("class()\nDuplicate interfaces.")       
      end if
      if ClassType[temp]!=INTERFACE then
         fatal_error("class()\nClass "&get_class_name(temp)
         &" in interface list is not an interface.")   
      end if
   end for           
   temp={ClassCounter}
   for i=1 to slen do
      temp&=SuperclassList[Extends[i]]    
   end for 
   Extends=NONE
   for i=1 to length(temp) do
      thisClass=temp[i]
      if not find(thisClass,Extends) then
         Extends&=thisClass   
      end if
   end for
   temp=NONE
   for i=1 to ilen do 
      temp&=SuperclassList[Implements[i]]   
   end for
   if slen>0 then
      temp&=InterfaceList[Extends[2]]      
   end if
   Implements=NONE
   for i=1 to length(temp) do                  
      thisClass=temp[i]
      if not find(thisClass,Implements) then
         Implements&=thisClass   
      end if
   end for  
   SuperclassList[CurrentDef]=Extends
   InterfaceList[CurrentDef]=Implements 
   if slen>0 then
      superClass=Extends[2]
      InnerClassList[CurrentDef]=InnerClassList[superClass]     
      PropertyName[CurrentDef]=PropertyName[superClass]
      PropertySource[CurrentDef]=PropertySource[superClass]
      PropertyReadAccess[CurrentDef]=PropertyReadAccess[superClass] 
      PropertyGetter[CurrentDef]=PropertyGetter[superClass]
      PropertyWriteAccess[CurrentDef]=PropertyWriteAccess[superClass]                                
      PropertySetter[CurrentDef]=PropertySetter[superClass]
      PropertyPersistence[CurrentDef]=PropertyPersistence[superClass]
      PropertyDefault[CurrentDef]=PropertyDefault[superClass]
      MethodName[CurrentDef]=MethodName[superClass]
      MethodSource[CurrentDef]=MethodSource[superClass]
      MethodType[CurrentDef]=MethodType[superClass]
      MethodAccess[CurrentDef]=MethodAccess[superClass]
      MethodID[CurrentDef]=MethodID[superClass]
      MethodDefault[CurrentDef]=MethodDefault[superClass]
      EventName[CurrentDef]=EventName[superClass]
      EventSource[CurrentDef]=EventSource[superClass]
      EventLinkAccess[CurrentDef]=EventLinkAccess[superClass]
      EventLinker[CurrentDef]=EventLinker[superClass]
      EventRaiseAccess[CurrentDef]=EventRaiseAccess[superClass]
      EventRaiser[CurrentDef]=EventRaiser[superClass]   
      if slen>1 then
         if Type=INTERFACE then
            merge_interface()
         end if
      end if
   end if   
   if innerFlag=TRUE then 
      OuterClassList[CurrentDef]&=outerClass&OuterClassList[outerClass] 
   end if   
   if Type<INTERFACE then 
      subset_check(superClass,CurrentDef)
      temp=OuterClassList[CurrentDef]     
      for i=1 to length(temp) do
         InnerClassList[temp[i]]&=CurrentDef  
      end for
   end if
   return This[ProgramCounter]
end function

global procedure end_class() 
   integer classType,index,int,thisClass
   sequence accessList,intList,intMeths,meth,methList,outerList,temp
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("end_class()\nNot allowed.")
   end if           
   classType=ClassType[CurrentDef]
   if classType<ABSTRACT then
      for i=INSTANCE to CLASS do
         methList=MethodType[CurrentDef][i] 
         for j=1 to length(methList) do
            if methList[j]=ABSTRACT then
               fatal_error("end_class()\nDoes not override "
               &get_class_name(MethodSource[CurrentDef][i][j])
               &" abstract "&CONSTANT_NAME[i]&" method "
               &MethodName[CurrentDef][i][j]&".")
            end if
         end for  
      end for   
   end if
   intList=InterfaceList[CurrentDef]
   for i=1 to length(intList) do
      int=intList[i]
      for j=INSTANCE to CLASS do
         methList=MethodName[CurrentDef][j]
         intMeths=MethodName[int][j]
         for k=1 to length(intMeths) do 
            meth=intMeths[k]
            index=find(meth,methList)
            if index=NIL or MethodAccess[CurrentDef][j][index]!=PUBLIC then
               fatal_error("end_class()\nDoes not implement "
               &get_class_name(MethodSource[int][j][k])
               &" interface "&CONSTANT_NAME[j]&" method "&meth
               &" as a public method.")      
            end if
         end for
      end for 
   end for 
   if classType<INTERFACE then
      accessList=SuperclassList[CurrentDef] 
      outerList=OuterClassList[CurrentDef]
      for i=1 to length(outerList) do       
         temp=SuperclassList[outerList[i]]
         for j=1 to length(temp) do
            thisClass=temp[j] 
            if not find(thisClass,accessList) then
               accessList&=thisClass    
            end if
         end for 
      end for  
      AccessList[CurrentDef]=accessList
   end if
   TempMethods=TempMethodsStack[1]
   TempMethodsStack=TempMethodsStack[2..length(TempMethodsStack)]
   PrototypeInstance[CurrentDef]=PropertyDefault[CurrentDef][INSTANCE]
   &repeat({DEFAULT_HANDLER},length(EventName[CurrentDef][INSTANCE]))
   ClassEntity[CurrentDef]=PropertyDefault[CurrentDef][CLASS]
   &repeat({DEFAULT_HANDLER},length(EventName[CurrentDef][CLASS]))
   FullyDefined[CurrentDef]=TRUE 
   VOID=system_call(This[ProgramCounter],"Initialize",NONE)  
   ProgramCounter-=1
   CurrentDef=Caller[ProgramCounter]  
end procedure

global function simple_class(object Name,object Type,object Extends,
object Implements) 
   sequence handle
   handle=class(Name,Type,Extends,Implements)
   end_class()
   return handle
end function

global procedure property(object Name,object entityType,object readAccess,
object writeAccess,object persistence,object default)
   integer classType,getter,index,setter 
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("property()\nNot allowed.")
   end if              
   classType=ClassType[CurrentDef]
   if classType>ABSTRACT then
      fatal_error("property()\nNot allowed in "&CONSTANT_CAP_NAME[classType]
      &" class.")
   end if
   if not identifier(Name) then
      fatal_error("property()\nInvalid property name.")
   end if
   if not find(entityType,VALID_ENTITY_TYPE) then
      fatal_error("property()\nProperty "&Name
      &" has invalid entity type.")
   end if            
   index=find(Name,PropertyName[CurrentDef][entityType])
   if index!=NIL then
      if PropertySource[CurrentDef][entityType][index]=CurrentDef then
         fatal_error("property()\n"&CONSTANT_CAP_NAME[entityType]
         &" property "&Name&" has already been defined.")
      end if
      fatal_error("property()\nAttempt to override "
      &get_class_name(PropertySource[CurrentDef][entityType][index])
      &" "&CONSTANT_NAME[entityType]&" property "&Name&".")
   end if
   if not accessor_specifier(readAccess) then
      fatal_error("property()\n"&CONSTANT_CAP_NAME[entityType]
      &" property "&Name&" has invalid read access.")
   end if                  
   if sequence(readAccess) then
      getter=readAccess[2]
      readAccess=readAccess[1]
   else
      getter=NULL_METHOD   
   end if 
   if not accessor_specifier(writeAccess) then
      fatal_error("property()\n"&CONSTANT_CAP_NAME[entityType]
      &" property "&Name&" has invalid write access.")
   end if     
   if sequence(writeAccess) then
      setter=writeAccess[2]
      writeAccess=writeAccess[1]
   else
      setter=NULL_METHOD    
   end if  
   if not find(persistence,VALID_PERSISTENCE) then
      fatal_error("property()\n"&CONSTANT_CAP_NAME[entityType]
      &" property "&Name&" has invalid persistence.") 
   end if
   PropertyName[CurrentDef][entityType]&={Name}
   PropertySource[CurrentDef][entityType]&=CurrentDef
   PropertyReadAccess[CurrentDef][entityType]&=readAccess
   PropertyGetter[CurrentDef][entityType]&=getter
   PropertyWriteAccess[CurrentDef][entityType]&=writeAccess
   PropertySetter[CurrentDef][entityType]&=setter
   PropertyPersistence[CurrentDef][entityType]&=persistence
   PropertyDefault[CurrentDef][entityType]&={default}
end procedure

global procedure method(object Name,object params,object entityType,
object methodType,object access,object ID,object default)
   integer classType,index
   sequence methType,methSource,methAccess,methID
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("method()\nNot allowed.")
   end if       
   classType=ClassType[CurrentDef]
   if classType=EXCEPTION then
      fatal_error("method()\nNot allowed in exception class.")   
   end if
   if not identifier(Name) then
      fatal_error("method()\nInvalid method name.")  
   end if
   if not integer(params) or params<0 then
      fatal_error("method()\nMethod "&Name
      &" has invalid parameter count.")
   end if
   Name&=sprintf("#%d",params)     
   if not find(entityType,VALID_ENTITY_TYPE) then
      fatal_error("method()\nMethod "&Name&" has invalid entity type.")
   end if
   index=find(Name,TempMethods[entityType])
   if index!=NIL then
      fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
      &" method "&Name&" has already been defined.")
   end if   
   TempMethods[entityType]&={Name}
   index=find(Name,MethodName[CurrentDef][entityType])
   if not find(methodType,VALID_METHOD_TYPE) then
      fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
      &" method "&Name&" has invalid method type.")
   end if 
   if not find(methodType,METHOD_TYPE_MATRIX[classType]) then
      fatal_error("method()\n"&CONSTANT_CAP_NAME[methodType]&" "
      &CONSTANT_NAME[entityType]&" method "&Name&" not allowed in "
      &CONSTANT_NAME[classType]&" class.")
   end if
   if not find(access,VALID_ACCESS) then
      fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
      &" method "&Name&" has invalid access.")
   end if  
   if access!=PUBLIC and classType=INTERFACE then
      fatal_error("method()\n"&CONSTANT_CAP_NAME[access]&" "
      &CONSTANT_NAME[entityType]&" method "&Name&" not allowed in "
      &"interface class.")   
   end if
   if not integer(ID) or ID=BAD_METHOD or ID<ABSTRACT_METHOD then
      fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
      &" method "&Name&" has invalid routine id.")
   end if
   if methodType=ABSTRACT then
      if ID!=ABSTRACT_METHOD then
         fatal_error("method()\nAbstract "&CONSTANT_NAME[entityType]
         &" method "&Name&" must have ABSTRACT_METHOD as routine id.")
      end if
   else
      if ID=ABSTRACT_METHOD then
         fatal_error("method()\n"&CONSTANT_CAP_NAME[methodType]
         &" "&CONSTANT_NAME[entityType]&" method "&Name
         &" must not have ABSTRACT_METHOD as routine id.")
      end if
   end if
   if atom(default) then 
      fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
      &" method "&Name&" has an atom as default parameter list.")
   end if 
   if ID=NULL_METHOD then
      if equal(default,NONE) then default={NIL} end if
      if length(default)!=1 then
         fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
         &" method "&Name&" has NULL_METHOD as routine_id. "
         &"Only one default value is allowed.")   
      end if
   end if   
   if ID=SUPER_METHOD then
      if length(default)!=0 then
         fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
         &" method "&Name&" has SUPER_METHOD as routine_id. "
         &"No default value is allowed.")      
      end if
   end if  
   if ID=ABSTRACT_METHOD then
      if length(default)!=0 then
         fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
         &" method "&Name&" has ABSTRACT_METHOD as routine_id. "
         &"No default value is allowed.")      
      end if
   end if
   if index=NIL then
      if ID=SUPER_METHOD then 
         fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
         &" method "&Name&" does not override a superclass method. "
         &"SUPER_METHOD may not be used as routine id.")
      end if
      MethodName[CurrentDef][entityType]&={Name}
      MethodSource[CurrentDef][entityType]&=CurrentDef
      MethodType[CurrentDef][entityType]&=methodType
      MethodAccess[CurrentDef][entityType]&=access
      MethodID[CurrentDef][entityType]&=ID
      MethodDefault[CurrentDef][entityType]&={default}
   else
      methSource=MethodSource[CurrentDef][entityType]
      methType=MethodType[CurrentDef][entityType]
      methAccess=MethodAccess[CurrentDef][entityType]
      methID=MethodID[CurrentDef][entityType]
      if not find(methodType,OVERRIDE_MATRIX[methType[index]]) then
         fatal_error("method()\n"&CONSTANT_CAP_NAME[methodType]&" "
         &CONSTANT_NAME[entityType]&" method "&Name
         &" not allowed to override superclass "
         &get_class_name(methSource[index])&" "
         &CONSTANT_NAME[methType[index]]&" method.")
      end if
      if access>methAccess[index] then
         fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
         &" method "&Name&" has more restrictive access than superclass "
         &get_class_name(methSource[index])&" method.")
      end if 
      if ID=SUPER_METHOD then 
         if methAccess[index]=PRIVATE or methType[index]=ABSTRACT then 
            fatal_error("method()\n"&CONSTANT_CAP_NAME[entityType]
            &" method "&Name&" overrides a private or abstract superclass "
            &get_class_name(methSource[index])&" method. "
            &"SUPER_METHOD may not be used as routine id.")
         end if
      end if
      if ID!=SUPER_METHOD then
         MethodSource[CurrentDef][entityType][index]=CurrentDef
         MethodID[CurrentDef][entityType][index]=ID 
         if ID!=ABSTRACT_METHOD then
            MethodDefault[CurrentDef][entityType][index]=default   
         end if
      end if
      MethodType[CurrentDef][entityType][index]=methodType
      MethodAccess[CurrentDef][entityType][index]=access  
   end if
end procedure    

global procedure event(object Name,object params,object entityType,
object linkAccess,object raiseAccess)
   integer classType,index,linker,raiser 
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("event()\nNot allowed.")
   end if           
   classType=ClassType[CurrentDef]
   if classType>ABSTRACT then
      fatal_error("event()\nNot allowed in "&CONSTANT_CAP_NAME[classType]
      &" class.")
   end if
   if not identifier(Name) then
      fatal_error("event()\nInvalid event name.")
   end if   
   if not find(entityType,VALID_ENTITY_TYPE) then
      fatal_error("event()\nEvent "&Name&" has invalid entity type.")
   end if  
   if not integer(params) or params<0 then
      fatal_error("event()\nEvent "&Name&" has invalid parameter count.")
   end if
   Name&=sprintf("#%d",params) 
   index=find(Name,EventName[CurrentDef][entityType])
   if index!=NIL then
      if EventSource[CurrentDef][entityType][index]=CurrentDef then
         fatal_error("event()\nInstance event "&Name
         &" has already been defined.")
      end if
      fatal_error("event()\nAttempt to override "
      &get_class_name(EventSource[CurrentDef][entityType][index])
      &" instance event "&Name&".")
   end if
   if not accessor_specifier(linkAccess) then
      fatal_error("event()\n"&CONSTANT_CAP_NAME[entityType]
      &" event "&Name&" has invalid link access.")
   end if                        
   if sequence(linkAccess) then
      linker=linkAccess[2]
      linkAccess=linkAccess[1]
   else
      linker=NULL_METHOD   
   end if
   if not accessor_specifier(raiseAccess) then
      fatal_error("event()\n"&CONSTANT_CAP_NAME[entityType]
      &" event "&Name&" has invalid raise access.")
   end if                             
   if sequence(raiseAccess) then
      raiser=raiseAccess[2]
      raiseAccess=raiseAccess[1]
   else
      raiser=NULL_METHOD   
   end if
   EventName[CurrentDef][entityType]&={Name}
   EventSource[CurrentDef][entityType]&=CurrentDef
   EventLinkAccess[CurrentDef][entityType]&=linkAccess
   EventLinker[CurrentDef][entityType]&=linker
   EventRaiseAccess[CurrentDef][entityType]&=raiseAccess
   EventRaiser[CurrentDef][entityType]&=raiser
end procedure   

global function interface(object Name,object Extends,
object instanceMethods,object classMethods)
   integer index,len,params,targetType
   sequence handle,methods 
   object meth
   handle=class(Name,INTERFACE,Extends,NONE)
   if atom(instanceMethods) or method_specifier(instanceMethods) 
   or method_name(instanceMethods) then
      instanceMethods={instanceMethods}
   end if
   len=length(instanceMethods)
   methods=instanceMethods&classMethods
   targetType=INSTANCE
   for i=1 to length(methods) do
      if i>len then targetType=CLASS end if
      meth=methods[i]     
      if method_name(meth) then
         index=find('#',meth) 
         params=0 
         for j=index+1 to length(meth) do
            params=params*10+meth[j]-'0'   
         end for   
         meth=meth[1..index-1]   
      elsif method_specifier(meth) then
         params=meth[2]
         meth=meth[1]  
      else
         params=0 
      end if
      method(meth,params,targetType,ABSTRACT,PUBLIC,ABSTRACT_METHOD,NONE)
   end for 
   end_class()
   return handle
end function

global function exception(object Name,object Extends)
   sequence handle
   handle=class(Name,EXCEPTION,Extends,NONE)      
   end_class()
   return handle
end function  

-- Exception handling:
global procedure throw(object theException)
   integer thisClass
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("throw()\nNot allowed.")
   end if
   if not class_entity(theException) then
      fatal_error("throw()\nInvalid exception.")
   end if
   thisClass=theException[1]
   if ClassType[thisClass]!=EXCEPTION then
      fatal_error("throw()\nClass "&get_class_name(thisClass)
      &" is not an exception.")
   end if  
   if PendingException!=NIL then 
      if thisClass=Type_Check_Failure_ID then
         if PendingException=Type_Check_Failure_ID then
            return   
         end if
      end if
      fatal_error("throw()\nException "&get_class_name(thisClass)
      &" thrown.")
   end if
   PendingException=thisClass
   CaughtException=NIL
end procedure  

global function catch(object theException)
   integer thisClass
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("catch()\nNot allowed.")
   end if
   if not class_entity(theException) then
      fatal_error("catch()\nInvalid exception.")
   end if
   thisClass=theException[1]
   if ClassType[thisClass]!=EXCEPTION then
      fatal_error("catch()\nClass "&get_class_name(thisClass)
      &" is not an exception.")
   end if
   if PendingException=NIL then return FALSE end if
   if not find(thisClass,SuperclassList[PendingException]) then
      return FALSE
   end if
   CaughtException=PendingException
   PendingException=NIL
   return TRUE
end function

global function caught()
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("caught()\nNot allowed.")
   end if
   if CaughtException=NIL then
      return Null_Class
   end if
   return {CaughtException,0,MARKER}
end function

global function success()
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("success()\nNot allowed.")
   end if
   return PendingException=NIL
end function 

-- Serialization:
-- Byte codes:      
-- 0 to 247 = immediate value
-- 248 = 2 byte integer
-- 249 = 3 byte integer
-- 250 = 4 byte integer
-- 251 = 4 byte atom
-- 252 = 8 byte atom
-- 253 = sequence (second byte is sequence type)
-- 254 = entity  
-- 255 = pointer 

-- 2 byte range = -power(2,15) to power(2,15)-1
-- 3 byte range = -power(2,23) to power(2,23)-1 
 

procedure clear_serialization()
   ByteSequence=NONE
   PackageList=NONE
   ClassList=NONE 
   EntityList=NONE 
   DSPointer=1 
   CloneOK=TRUE
   BadClone=FALSE  
end procedure

type byte_sequence(object x) 
   object byte
   if atom(x) then return FALSE end if
   for i=1 to length(x) do
      byte=x[i]
      if not integer(byte) then return FALSE end if
      if byte<0 then return FALSE end if
      if byte>255 then return FALSE end if
   end for 
   return TRUE
end type

constant CS_MULT={1,2,3,4,4,8}

function classify_sequence(sequence seq)
   integer len,genlen
   sequence temp,typeseq
   object x
   len=length(seq)
   typeseq=repeat(len,6)*CS_MULT
   genlen=0
   for i=1 to len do
      x=seq[i]
      if sequence(x) then return 253 end if 
      if integer(x) then
         if x>=0 and x<=247 then 
            genlen+=1       
         elsif x>=248 and x<=255 then
            genlen+=3
         elsif x>=-MAX_2B and x<MAX_2B then 
            typeseq[1]=-1 
            genlen+=3
         elsif x>=-MAX_3B and x<=MAX_3B then 
            typeseq[1..2]=-1 
            genlen+=4
         else
            typeseq[1..3]=-1
            genlen+=5
         end if
      else
         temp=machine_func(48,x) -- atom_to_float_32
         if x=machine_func(49,temp) then -- float32_to_atom
            typeseq[1..4]=-1
            genlen+=5
         else
            typeseq[1..5]=-1
            genlen+=9
         end if
      end if
   end for
   for i=1 to 6 do
      x=typeseq[i]
      if x!=-1 then
         if x<=genlen then
            return 246+i
         end if
      end if
   end for 
   return 253
end function

procedure serialize(object x) 
   integer classIndex,index,len,stype,targetType,thisClass,thisInstance
   integer thisPackage
   sequence propName,propPersist,temp
   object propValue,xi
   if atom(x) then
      if integer(x) then
         if x>=0 and x<=247 then
            ByteSequence&=x
         else
            poke4(MEM,x)
            if x>=-MAX_2B and x<MAX_2B then
               ByteSequence&=248&peek({MEM,2})
            elsif x>=-MAX_3B and x<MAX_3B then
               ByteSequence&=249&peek({MEM,3})
            else
               ByteSequence&=250&peek({MEM,4})
            end if
         end if 
      else
         temp=machine_func(48,x) -- atom_to_float32
         if x=machine_func(49,temp) then  -- float32_to_atom
            ByteSequence&=251&temp
         else
            ByteSequence&=252&machine_func(46,x) -- atom_to_float64
         end if
      end if
   elsif entity(x) then
      index=find(x,EntityList)
      if index=NIL then 
         EntityList&={x}
         thisClass=x[1]
         thisInstance=x[2] 
         targetType=(thisInstance=0)+1
         classIndex=find(thisClass,ClassList)
         if classIndex=NIL then
            ClassList&=thisClass 
            classIndex=length(ClassList)
         end if
         thisPackage=ClassName[thisClass][2]
         index=find(thisPackage,PackageList)
         if index=NIL then PackageList&=thisPackage end if
         ByteSequence&=254
         serialize(classIndex) 
         if targetType=INSTANCE then 
            ByteSequence&=1 
            propValue=InstanceEntity[thisInstance]   
         else
            ByteSequence&=0 
            propValue=ClassEntity[thisClass] 
         end if
         propName=PropertyName[thisClass][targetType]
         propPersist=PropertyPersistence[thisClass][targetType] 
         if atom(propValue) then
            ByteSequence&=1
            return   
         end if
         VOID=system_call(x,"Before_Save",NONE)
         for i=1 to length(propName) do
            if propPersist[i]=PERSISTENT then
               serialize(propName[i])
               serialize(propValue[i])
            end if  
         end for                    
         ByteSequence&=0         
         VOID=system_call(x,"After_Save",NONE)
      else
         ByteSequence&=255
         serialize(index)   
      end if
   else
      ByteSequence&=253                
      len=length(x)
      if len=0 then 
         ByteSequence&=0
         return   
      end if
      stype=classify_sequence(x)
      ByteSequence&=stype
      serialize(len)
      if stype=247 then
         ByteSequence&=x
         return     
      end if
      for i=1 to len do
         xi=x[i]
         if stype<251 then
            poke4(MEM,xi)
            ByteSequence&=peek({MEM,stype-246})
         elsif stype=251 then
            ByteSequence&=machine_func(48,xi) -- atom_to_float32
         elsif stype=252 then
            ByteSequence&=machine_func(46,xi) -- atom_to_float64
         else
            serialize(xi)
         end if
      end for
   end if
end procedure

global function save_entity(object theEntity) 
   integer index,len,pad,thisClass
   sequence byteseq,check,className 
   CaughtException=NIL          
   if Context[ProgramCounter]=CLASS_DEFINITION or PendingException!=NIL then
      fatal_error("save_entity()\nNot allowed.")
   end if
   if length(EntityList)!=0 then
      fatal_error("save_entity()\nAn entity is currently being saved "
      &"or restored.")   
   end if
   if not entity(theEntity) then 
      fatal_error("save_entity()\nParameter is not an entity.")
   end if        
   serialize(theEntity)              
   byteseq=ByteSequence
   ByteSequence=NONE 
   serialize(SESSION_ID)
   for i=1 to length(PackageList) do
      serialize(PackageName[PackageList[i]])
   end for                     
   ByteSequence&=0
   for i=1 to length(ClassList) do
      thisClass=ClassList[i] 
      className=ClassName[thisClass]
      index=find(className[2],PackageList)
      serialize(index)
      serialize(className[1])   
   end for        
   ByteSequence&=0
   byteseq=CURRENT_VERSION&ByteSequence&byteseq
   clear_serialization()
   len=length(byteseq)
   pad=4-remainder(len,4) 
   byteseq&=repeat(pad,pad)
   len+=pad
   check=repeat(0,4)
   for i=1 to len/2 do
      check=remainder(check+{byteseq[i],byteseq[len/2+i],byteseq[2*i-1],
      byteseq[2*i]},256)
   end for 
   byteseq&=check
   for i=2 to len+4 do
      byteseq[i]=xor_bits(byteseq[i],byteseq[i-1])
   end for
   return byteseq
end function  

function deserialize() 
   integer bad,deleted,index,key,len,stype,saveBadClone
   integer targetType,thisClass,thisInstance
   sequence propName,propPersist,theEntity
   object name,propValue,val 
   key=ByteSequence[DSPointer]
   DSPointer+=1
   if key<=247 then
      return key
   elsif key=248 then
      poke(MEM,ByteSequence[DSPointer..DSPointer+1])
      if ByteSequence[DSPointer+1]>127 then
         poke(MEM+2,{255,255})   
      else
         poke(MEM+2,{0,0})
      end if                
      DSPointer+=2
      return peek4s(MEM)  
   elsif key=249 then
      poke(MEM,ByteSequence[DSPointer..DSPointer+2])
      if ByteSequence[DSPointer+2]>127 then
         poke(MEM+3,255)
      else
         poke(MEM+3,0)
      end if          
      DSPointer+=3
      return peek4s(MEM)
   elsif key=250 then
      poke(MEM,ByteSequence[DSPointer..DSPointer+3])
      DSPointer+=4
      return peek4s(MEM)
   elsif key=251 then
      val=machine_func(49,ByteSequence[DSPointer..DSPointer+3])
      -- float32_to_atom
      DSPointer+=4
      return val
   elsif key=252 then
      val=machine_func(47,ByteSequence[DSPointer..DSPointer+7])
      -- float64_to_atom
      DSPointer+=8
      return val
   elsif key=253 then
      stype=ByteSequence[DSPointer]
      DSPointer+=1
      if stype=0 then return NONE end if
      len=deserialize()
      if stype=247 then
         val=ByteSequence[DSPointer..DSPointer+len-1]
         DSPointer+=len
         return val   
      end if  
      val=NONE
      for i=1 to len do
         if stype!=253 then
            DSPointer-=1
            ByteSequence[DSPointer]=stype
         end if
         val&={deserialize()}   
      end for 
      return val
   elsif key=254 then
      thisClass=ClassList[deserialize()]
      if deserialize()!=0 then 
         targetType=INSTANCE 
         propValue=PrototypeInstance[thisClass]
         if thisClass=3 then
            thisInstance=1
         else
            InstanceCounter+=1
            if InstanceCounter>length(InstanceEntity) then
               InstanceEntity&=repeat(NIL,length(InstanceEntity))
            end if                                     
            thisInstance=InstanceCounter
         end if
      else    
         targetType=CLASS
         propValue=ClassEntity[thisClass]        
         thisInstance=0  
      end if    
      propName=PropertyName[thisClass][targetType]
      propPersist=PropertyPersistence[thisClass][targetType]
      saveBadClone=BadClone
      if CloneOK=FALSE and find(5,InterfaceList[thisClass]) then
         BadClone=TRUE   
      end if 
      if BadClone=TRUE then
         if targetType=INSTANCE then
            theEntity=Null_Instance
            InstanceCounter-=1
         else
            theEntity=Null_Class
         end if
         EntityList&={theEntity}  
      else   
         theEntity={thisClass,thisInstance,MARKER}
         EntityList&={theEntity} 
         if targetType=CLASS then
            VOID=system_call(theEntity,"Before_Restore",NONE)
         end if
      end if
      bad=BadClone
      name=deserialize()
      deleted=equal(name,1)
      while sequence(name) do
         val=deserialize()
         index=find(name,propName)
         if index!=NIL then
            if propPersist[index]=PERSISTENT then
               propValue[index]=val   
            end if
         end if
         name=deserialize()
      end while
      if deleted=TRUE then propValue=NIL end if
      if bad=FALSE then
         if targetType=INSTANCE then
            InstanceEntity[thisInstance]=propValue   
         else
            ClassEntity[thisClass]=propValue   
         end if
         if deleted=FALSE then
            VOID=system_call(theEntity,"After_Restore",NONE)
         end if
      end if
      BadClone=saveBadClone           
      return theEntity
   else
      val=deserialize()     
      return EntityList[val]   
   end if
end function

global function restore_entity(object byteseq)
   integer index,len                                
   sequence check,theEntity,version
   object x  
   CaughtException=NIL                                 
   if Context[ProgramCounter]=CLASS_DEFINITION or PendingException!=NIL then
      fatal_error("restore_entity()\nNot allowed.")
   end if 
   if length(EntityList)!=0 then
      fatal_error("restore_entity()\nAn entity is currently being saved "
      &"or restored.")   
   end if
   if not byte_sequence(byteseq) then return Null_Instance end if    
   len=length(byteseq)
   if remainder(len,4)!=0 or len<17 then return Null_Instance end if
   for i=len to 2 by -1 do
      byteseq[i]=xor_bits(byteseq[i],byteseq[i-1])   
   end for
   version=byteseq[1..3]
   if not find(version,VERSION_LIST) then return Null_Instance end if
   check=repeat(0,4)
   len-=4
   for i=1 to len/2 do
      check=remainder(check+{byteseq[i],byteseq[len/2+i],byteseq[2*i-1],
      byteseq[2*i]},256)       
   end for 
   if not equal(check,byteseq[len+1..len+4]) then return Null_Instance end if
   ByteSequence=byteseq[4..len-byteseq[len]] 
   if equal(deserialize(),SESSION_ID) then 
      CloneOK=FALSE
   else
      CloneOK=TRUE
   end if
   while TRUE do
      x=deserialize()
      if atom(x) then exit end if 
      index=find(x,PackageName)
      if index=NIL then 
         clear_serialization()
         return Null_Instance
      end if 
      PackageList&=index  
   end while 
   while TRUE do
      x=deserialize()
      if x=0 then exit end if
      index=PackageList[x]
      x=deserialize()
      index=find({x,index},ClassName)
      if index=NIL then
         clear_serialization()
         return Null_Instance   
      end if 
      ClassList&=index
   end while
   theEntity=deserialize()
   clear_serialization()
   return theEntity
end function      

                                           
-- Predefined class methods: 
package("Diamond Kernel") 

ClassCounter=1 
CurrentDef=1 
ProgramCounter=2
Context[2]=CLASS_DEFINITION
Caller[2]=1
This[2]=Entity -- class()

function Entity_new_0()
   integer thisClass,classType
   sequence thisEntity
   thisEntity=This[ProgramCounter]
   thisClass=thisEntity[1]
   classType=ClassType[thisClass]
   if classType=ABSTRACT then return Null_Instance end if
   InstanceCounter+=1
   if InstanceCounter>length(InstanceEntity) then
      InstanceEntity&=repeat(NIL,length(InstanceEntity))
   end if
   InstanceEntity[InstanceCounter]=PrototypeInstance[thisClass]
   thisEntity[2]=InstanceCounter
   return thisEntity
end function    

method("new",0,CLASS,NORMAL,PROTECTED,routine_id("Entity_new_0"),NONE) 

function Entity_clone_0() 
   integer thisClass
   sequence thisEntity
   thisEntity=This[ProgramCounter]
   thisClass=thisEntity[1]
   if find(Do_Not_Clone_ID,InterfaceList[thisClass]) then 
      return Null_Instance
   end if
   InstanceCounter+=1
   if InstanceCounter>length(InstanceEntity) then
      InstanceEntity&=repeat(NIL,1024)
   end if
   InstanceEntity[InstanceCounter]=InstanceEntity[thisEntity[2]]
   thisEntity[2]=InstanceCounter
   return thisEntity
end function

method("clone",0,INSTANCE,NORMAL,PROTECTED,routine_id("Entity_clone_0"),NONE)

function Entity_delete_0()
   integer thisInstance
   thisInstance=This[ProgramCounter][2]
   InstanceEntity[thisInstance]=NIL
   return Null_Instance
end function

method("delete",0,INSTANCE,NORMAL,PROTECTED,
routine_id("Entity_delete_0"),NONE)  

TempMethods=DOUBLE_NONE
ProgramCounter=1
ClassCounter=3  
CurrentDef=NIL  


VOID=interface("Clonable",NONE,"clone#0",NONE)

VOID=interface("Do_Not_Clone",NONE,NONE,NONE)

VOID=exception("Type_Check_Failure",Exception)

package("Diamond Default")     

-- Exception throwing type checking

constant BOOLEAN={0,1}  

global type boolean(object x)
   return find(x,BOOLEAN)!=NIL   
end type

global type is_boolean(object x)
   if boolean(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type

global type is_integer(object x)
   if integer(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type                      

global type is_atom(object x)
   if atom(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type                      

global type is_sequence(object x) 
   if sequence(x) then 
      if not entity(x) then return TRUE end if
   end if
   throw(Type_Check_Failure)
   return TRUE
end type 

global type is_object(object x)
   if not entity(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type

global type is_identifier(object x)
   if identifier(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type

global type is_entity(object x)
   if entity(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE   
end type                            
                                                                 
global type is_instance(object x)
   if instance_entity(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type                         

global type is_live_instance(object x)
   if live_instance(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type

global type is_class(object x)
   if class_entity(x) then return TRUE end if
   throw(Type_Check_Failure)
   return TRUE
end type  

global function do_type_check(object x,integer ID)
   object result
   result=call_func(ID,{x})
   if atom(result) then
      if result!=0 then return TRUE end if
   end if
   throw(Type_Check_Failure)
   return TRUE
end function  

global function live_instance_check(object x,object c)         
   if not class_entity(c) then
      fatal_error("live_instance_check()\nSecond parameter is not a class.")   
   end if  
   while TRUE do
      if not instance_entity(x) then exit end if
      if atom(InstanceEntity[x[2]]) then exit end if
      if class_type(c)=INTERFACE then 
         if implements(x,c) then return TRUE end if
      else
         if extends(x,c) then return TRUE end if
      end if
      exit
   end while  
   throw(Type_Check_Failure)
   return TRUE
end function 

global type is_Clonable(object x)
   return live_instance_check(x,Clonable)   
end type                          

global type is_Do_Not_Clone(object x)
   return live_instance_check(x,Do_Not_Clone)   
end type  

