-- dvar.e v4.0.0 by Mike Nelson   
-- Diamond Standard Class Library
-- Reference variables, constants, and structures with events.   
-- IMPORTANT NOTE: Type checking will be lost if instances of these classes
-- are saved and restored,  since the Type_Check and Type_Checks
-- properties are transient. Because routine_id is used, restoring these 
-- properties would be unsafe among different applications
-- or different versions of the same application.     
 
without trace
without warning

include dcontain.e 


package("Diamond Standard Class Library") 


global constant Uninitialized_Variable=
exception("Uninitialized_Variable",Not_Allowed)  


constant NO_TYPE_CHECK=-2

type valid_type_check(object x)
   if integer(x) then
      if x>=NO_TYPE_CHECK then
         if x!=-1 then return TRUE end if
      end if
   end if
   throw(Type_Check_Failure)
   return TRUE
end type


global constant Variable=class("Variable",NORMAL,Entity,Clonable) 

global type is_Variable(object x)
   return live_instance_check(x,Variable)   
end type           

property("Data",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NIL)
property("Initialized",INSTANCE,PUBLIC,PRIVATE,PERSISTENT,FALSE)  
property("Type_Check",INSTANCE,PRIVATE,PRIVATE,TRANSIENT,NO_TYPE_CHECK)

function Variable_new_1(valid_type_check typeCheck)  
   entity new
   if not success() then return Null_Instance end if
   new=call_method(super(),"new",NONE)
   set_property(new,"Type_Check",typeCheck)
   return new
end function 
                
method("new",0,CLASS,NORMAL,PUBLIC,SUPER_METHOD,NONE)
method("new",1,CLASS,NORMAL,PUBLIC,routine_id("Variable_new_1"),NONE)

method("clone",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE)

method("delete",0,INSTANCE,NORMAL,PUBLIC,SUPER_METHOD,NONE) 

function Variable_Set_1(object new)
   integer typeCheck 
   object old
   typeCheck=get_property(this(),"Type_Check")
   if typeCheck!=NO_TYPE_CHECK then
      if call_func(typeCheck,{new})=FALSE then
         throw(Type_Check_Failure)   
      end if                      
      if not success() then return NIL end if
   end if       
   old=get_property(this(),"Data")
   set_property(this(),"Data",new)
   set_property(this(),"Initialized",TRUE)
   raise_event(this(),"Set",{old,new})
   return NIL
end function

method("Set",1,INSTANCE,NORMAL,PUBLIC,routine_id("Variable_Set_1"),NONE)

function Variable_Get_0()  
   object val
   if get_property(this(),"Initialized")=FALSE then
      throw(Uninitialized_Variable)
      return NIL   
   end if                
   val=get_property(this(),"Data")
   raise_event(this(),"Get",{val})
   return val   
end function  

method("Get",0,INSTANCE,NORMAL,PUBLIC,routine_id("Variable_Get_0"),NONE)  

function Variable_Uninitialize_1(valid_type_check ID)
   object old 
   if not success() then return NIL end if
   old=get_property(this(),"Data")
   set_property(this(),"Data",NIL)
   set_property(this(),"Initialized",FALSE)
   set_property(this(),"Type_Check",ID)
   return old   
end function    

method("Uninitialize",0,INSTANCE,FINAL,PUBLIC,
routine_id("Variable_Uninitialize_1"),{NO_TYPE_CHECK})
method("Uninitialize",1,INSTANCE,FINAL,PUBLIC,
routine_id("Variable_Uninitialize_1"),NONE)

event("Set",2,INSTANCE,PUBLIC,PRIVATE) 
event("Get",1,INSTANCE,PUBLIC,PRIVATE)

end_class()


global constant Initialized_Constant=
exception("Initialized_Constant",Not_Allowed)


global constant Constant=class("Constant",NORMAL,Variable,NONE) 

global type is_Constant(object x)
   return live_instance_check(x,Constant)   
end type           

function Constant_Set_1(object val)
   if get_property(this(),"Initialized")=TRUE then
      throw(Initialized_Constant)
      return NIL      
   end if       
   return call_method(super(),"Set",{val})
end function  

method("Set",1,INSTANCE,NORMAL,PUBLIC,routine_id("Constant_Set_1"),NONE)  

end_class()                     


type valid_handlers(object x)  
   integer len
   object xi
   if atom(x) then x={x} end if 
   len=length(x)
   for i=1 to len do
      xi=x[i] 
      if not integer(xi) then exit end if
      if xi<0 then 
         if not find(xi,{DEFAULT_HANDLER,NULL_HANDLER,CURRENT_HANDLERS}) then
            exit
         end if
      end if 
      if i=len then return TRUE end if
   end for
   throw(Type_Check_Failure)                                                        
   return TRUE
end type 


global constant Structure=class("Structure",NORMAL,Entity,Clonable)

global type is_Structure(object x)
   return live_instance_check(x,Structure)   
end type           
                                   
property("Elements",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)
property("Variables",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,NONE)
property("Type_Checks",INSTANCE,PRIVATE,PRIVATE,TRANSIENT,NONE)

method("new",0,CLASS,NORMAL,PUBLIC,SUPER_METHOD,NONE)

function Structure_clone_0()
   sequence variables
   entity clone
   clone=call_method(super(),clone,NONE)
   variables=get_property(clone,"Variables")
   for i=1 to length(variables) do
      variables[i]=call_method(variables[i],"clone",NONE)   
   end for                                     
   set_property(clone,"Variables",variables)
   return clone
end function

method("clone",0,INSTANCE,NORMAL,PUBLIC,routine_id("Structure_clone_0"),NONE)

function Structure_delete_0()
   sequence variables
   variables=get_property(this(),"Variables")
   for i=1 to length(variables) do
      VOID=call_method(variables[i],"delete",NONE)   
   end for                                   
   return call_method(super(),"delete",NONE)
end function                                

method("delete",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Structure_delete_0"),NONE)

function Structure_Add_Element_2(is_identifier name,valid_type_check ID)                
   sequence elements,typeChecks,variables
   if not success() then return NIL end if
   elements=get_property(this(),"Elements")
   if find(name,elements) then
      throw(Duplicate_Element)
      return NIL   
   end if       
   elements=append(elements,name) 
   set_property(this(),"Elements",elements) 
   variables=get_property(this(),"Variables") 
   variables=append(variables,call_method(Variable,"new",{ID}))
   set_property(this(),"Variables",variables)
   typeChecks=get_property(this(),"Type_Checks")&ID  
   set_property(this(),"Type_Checks",typeChecks)
   return NIL
end function 

method("Add_Element",1,INSTANCE,NORMAL,PUBLIC,
routine_id("Structure_Add_Element_2"),{NO_TYPE_CHECK})
method("Add_Element",2,INSTANCE,NORMAL,PUBLIC,
routine_id("Structure_Add_Element_2"),NONE)   

function Structure_Prototype_0()
   sequence elements,typeChecks,variables
   entity new           
   new=call_method(this_class(),"new",NONE)
   elements=get_property(this(),"Elements") 
   set_property(new,"Elements",elements)
   typeChecks=get_property(this(),"Type_Checks")
   set_property(new,"Type_Checks",typeChecks)
   variables=NONE
   for i=1 to length(elements) do
      variables=append(variables,call_method(Variable,"new",{typeChecks[i]}))   
   end for 
   set_property(new,"Variables",variables)
   return new
end function 

method("Prototype",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Structure_Prototype_0"),NONE)

function find_element(is_identifier name)
   integer index
   sequence elements,variables
   if not success() then return Null_Instance end if 
   elements=get_property(this(),"Elements")
   index=find(name,elements)
   if index=NIL then
      throw(Missing_Element)
      return Null_Instance
   end if 
   variables=get_property(this(),"Variables")
   return variables[index]    
end function

function Structure_Set_2(object name,object val)
   entity variable
   variable=find_element(name)
   if not success() then return NIL end if
   return call_method(variable,"Set",{val})
end function                              

method("Set",2,INSTANCE,NORMAL,PUBLIC,routine_id("Structure_Set_2"),NONE)

function Structure_Get_1(object name)
   entity variable
   variable=find_element(name)
   if not success() then return NIL end if
   return call_method(variable,"Get",NONE)
end function

method("Get",1,INSTANCE,NORMAL,PUBLIC,routine_id("Structure_Get_1"),NONE) 

function Structure_Link_Handler(object name,valid_handlers ID,object hname)
   entity variable  
   if not success() then return NIL end if
   variable=find_element(name)
   if not success() then return NIL end if
   link_handler(variable,hname,ID) 
   return NIL  
end function 
        
method("Link_Set_Handler",2,INSTANCE,FINAL,PUBLIC,
routine_id("Structure_Link_Handler"),{"Set#2"})        
             
method("Link_Get_Handler",2,INSTANCE,FINAL,PUBLIC,
routine_id("Structure_Link_Handler"),{"Get#1"})

function Structure_Initialized_0()
   sequence variables
   variables=get_property(this(),"Variables")
   for i=1 to length(variables) do
      if get_property(variables[i],"Initialized")=FALSE then 
         return FALSE
      end if
   end for                                                          
   return TRUE
end function  

method("Initialized",0,INSTANCE,FINAL,PUBLIC,
routine_id("Structure_Initialized_0"),NONE)

function Structure_Initialized_1(object name)
   entity variable
   variable=find_element(name)
   if catch(Exception) then return FALSE end if
   return get_property(variable,"Initialized")   
end function   

method("Initialized",1,INSTANCE,FINAL,PUBLIC,
routine_id("Structure_Initialized_1"),NONE) 

function Structure_Uninitialize_2(object name,object ID)
   entity variable                      
   variable=find_element(name)
   if not success() then return NIL end if
   return call_method(variable,"Uninitialize",{ID})
end function

method("Uninitialize",1,INSTANCE,FINAL,PUBLIC,
routine_id("Structure_Uninitialize_2"),{NO_TYPE_CHECK})
method("Uninitialize",2,INSTANCE,FINAL,PUBLIC,
routine_id("Structure_Uninitialize_2"),NONE)  

function Structure_After_Restore_0()
   sequence elements
   elements=get_property(this(),"Elements")
   set_property(this(),"Type_Checks",repeat(NO_TYPE_CHECK,length(elements)))
   return NIL   
end function  

method("After_Restore",0,INSTANCE,NORMAL,PROTECTED,
routine_id("Structure_After_Restore_0"),NONE)

end_class()            
           
            
package("Diamond Default") 
   


