-- dvsaver.e v4.0.0 by Mike Nelson   
-- Diamond Standard Class Library
-- Automates saving and restoring variables

without trace
without warning
                                 
include database.e
include dcontain.e  


function iff(atom condition,object isTrue,object isFalse)
   if condition then return isTrue else return isFalse end if
end function

constant SLASH=iff(platform()=LINUX,'/','\\')

constant LOCKING=iff(platform()=DOS32 and atom(getenv("windir")),
DB_LOCK_NO,DB_LOCK_EXCLUSIVE)

constant SAVE_DIR=getenv("EUDIR")&SLASH&"Dvsaver" 
constant SAVE_FILE=SAVE_DIR&SLASH&"dvsaver.edb"


package("Diamond Standard Class Library") 


global constant Var_Saver=class("Var_Saver",FINAL,Entity,NONE)

global type is_Var_Saver(object x)
   return live_instance_check(x,Var_Saver)   
end type

property("Table",INSTANCE,PRIVATE,PRIVATE,TRANSIENT,NONE)
property("Saved",INSTANCE,PRIVATE,PRIVATE,TRANSIENT,Null_Instance)
property("Restored",INSTANCE,PRIVATE,PRIVATE,TRANSIENT,Null_Instance)

function Var_Saver_new_1(is_identifier table)
   integer flag,openfile
   entity excep,new
   if not success() then return Null_Instance end if
   if atom(dir(SAVE_DIR)) then
      system("mkdir "&SAVE_DIR,2)
      if atom(dir(SAVE_DIR)) then 
         throw(File_Open_Error) 
         return Null_Instance
      end if
   end if                        
   if atom(dir(SAVE_FILE)) then
      flag=db_create(SAVE_FILE,LOCKING)
   else
      flag=db_open(SAVE_FILE,LOCKING)
   end if
   if flag=DB_OPEN_FAIL then
      throw(File_Open_Error)
      return Null_Instance   
   end if
   if flag=DB_LOCK_FAIL then
      db_close() 
      throw(File_Lock_Error)
      return Null_Instance
   end if
   if db_select_table(table)!=DB_OK then 
      if db_create_table(table)!=DB_OK then 
         throw(File_Write_Error)
         return Null_Instance
      end if
   end if
   db_close()
   new=call_method(super(),"new",NONE)
   set_property(new,"Table",table)
   set_property(new,"Saved",
   call_method(Indexed_Sequence,"new",{FALSE,FALSE,TRUE,1,FALSE}))
   return new
end function
 
method("new",1,CLASS,FINAL,PUBLIC,routine_id("Var_Saver_new_1"),NONE)  

function Var_Saver_delete_0()
   sequence restored
   VOID=call_method(get_property(this(),"Saved"),"delete",NONE)
   restored=get_property(this(),"Restored")
   if not equal(restored,Null_Instance) then
      VOID=call_method(restored,"delete",NONE)
   end if                                     
   return call_method(super(),"delete",NONE)
end function              

method("delete",0,INSTANCE,FINAL,PUBLIC,
routine_id("Var_Saver_delete_0"),NONE)
 
function Var_Saver_Save_2(object name,object val)
   integer index
   entity saved
   saved=get_property(this(),"Saved")
   index=call_method(saved,"Find",{name})
   if index>0 then
      VOID=call_method(saved,"Remove",{index})   
   end if
   VOID=call_method(saved,"Insert",{{name,val}})
   return NIL
end function

method("Save",2,INSTANCE,FINAL,PUBLIC,routine_id("Var_Saver_Save_2"),NONE)

function Var_Saver_Write_0() 
   integer flag
   sequence key,saved,table      
   entity excep
   table=get_property(this(),"Table")
   flag=db_open(SAVE_FILE,LOCKING)
   if flag=DB_OPEN_FAIL then 
      throw(File_Open_Error)
      return NIL
   end if
   if flag=DB_LOCK_FAIL then
      db_close()
      throw(File_Lock_Error)
      return NIL
   end if
   if db_select_table(table)!=DB_OK then
      if db_create_table(table)!=DB_OK then
         db_close()
         throw(File_Write_Error)
         return NIL
      end if
   end if             
   key=date()         
   saved=get_property(this(),"Saved")
   while TRUE do
      if db_insert(key,save_entity(saved))=DB_OK then exit end if
      key[5]+=.000001   
   end while
   db_close()
   return NIL
end function 

method("Write",0,INSTANCE,FINAL,PUBLIC,routine_id("Var_Saver_Write_0"),NONE)  

function Var_Saver_Clear_Saved_0()
   entity saved
   saved=get_property(this(),"Saved")
   for i=call_method(saved,"Count",NONE) to 1 by -1 do
      VOID=call_method(saved,"Remove",{i})    
   end for              
   return NIL
end function

method("Clear_Saved",0,INSTANCE,FINAL,PUBLIC,
routine_id("Var_Saver_Clear_Saved_0"),NONE)

function Var_Saver_Clear_Restored_0()
   sequence restored
   restored=get_property(this(),"Restored")
   if equal(restored,Null_Instance) then return NIL end if
   VOID=call_method(restored,"delete",NONE)
   set_property(this(),"Restored",Null_Instance)
   return NIL
end function

method("Clear_Restored",0,INSTANCE,FINAL,PUBLIC,
routine_id("Var_Saver_Clear_Restored_0"),NONE)

function Var_Saver_Read_0()
   integer flag,tableSize
   sequence restored,table
   entity excep 
   table=get_property(this(),"Table")
   flag=db_open(SAVE_FILE,LOCKING)
   if flag=DB_OPEN_FAIL then 
      throw(File_Open_Error)
      return NIL
   end if
   if flag=DB_LOCK_FAIL then
      db_close()
      throw(File_Lock_Error)
      return NIL   
   end if 
   if db_select_table(table)!=DB_OK then 
      db_close()
      throw(File_Read_Error)
      return NIL
   end if           
   tableSize=db_table_size()              
   if tableSize=0 then 
      db_close()      
      throw(File_Read_Error)
      return NIL
   end if
   restored=restore_entity(db_record_data(tableSize))       
   db_close()
   VOID=call_method(this(),"Clear_Restored",NONE)
   if equal(restored,Null_Instance) then 
      throw(Bad_Data)
      return NIL
   end if
   set_property(this(),"Restored",restored)
   return NIL
end function                               
                                                    
method("Read",0,INSTANCE,FINAL,PUBLIC,routine_id("Var_Saver_Read_0"),NONE)
   
function Var_Saver_Restore_1(object name)                
   integer index  
   sequence var
   entity restored
   restored=get_property(this(),"Restored")
   if equal(restored,Null_Instance) then 
      throw(Bad_Data)
      return NIL
   end if
   index=call_method(restored,"Find",{name})
   if index<0 then 
      throw(Type_Check_Failure)
      return NIL
   end if
   var=call_method(restored,"Get",{index})
   return var[1]
end function  
                           
method("Restore",1,INSTANCE,FINAL,PUBLIC,
routine_id("Var_Saver_Restore_1"),NONE)                           
                           
function Var_Saver_Clear_Table_1(object saveLast)  
   integer flag,start,stop,tableSize         
   sequence table
   entity excep
   table=get_property(this(),"Table")
   flag=db_open(SAVE_FILE,LOCKING)
   if flag=DB_OPEN_FAIL then 
      throw(File_Open_Error)
      return NIL
   end if
   if flag=DB_LOCK_FAIL then
      db_close()
      throw(File_Lock_Error)
      return NIL
   end if 
   if db_select_table(table)!=DB_OK then 
      db_close()
      throw(File_Read_Error)
      return NIL
   end if        
   tableSize=db_table_size()
   if not integer(saveLast) then saveLast=0 end if
   if saveLast>=0 then
      start=tableSize-saveLast
      stop=1
   else 
      start=tableSize   
      stop=tableSize+saveLast+1
      if stop<1 then stop=1 end if
   end if 
   for i=start to stop by -1 do
      db_delete_record(i)
   end for
   VOID=db_compress()
   if platform()=LINUX then
      system("rm "&SAVE_DIR&SLASH&"dvsaver.t*",2)
   else
      system("del "&SAVE_DIR&SLASH&"dvsaver.t*",2)
   end if
   db_close()
   return NIL
end function                                  

method("Clear_Table",0,INSTANCE,FINAL,PUBLIC,
routine_id("Var_Saver_Clear_Table_1"),{0})
method("Clear_Table",1,INSTANCE,FINAL,PUBLIC,
routine_id("Var_Saver_Clear_Table_1"),NONE)

end_class()

                      
package("Diamond Default")     


