-- dwidget.e v4.0.0 by Mike Nelson    
-- Diamond Standard Class Library
-- Visual Basic style components and collections.
-- Provides the functionality for building widgets.

without trace
without warning
                
include get.e                
include dcontain.e  


function validate_index(object index,integer len,integer plusOne)
   while TRUE do
      if not integer(index) then exit end if
      if index>len+plusOne then exit end if
      if index>0 then return index end if
      index=len+1+index  
      if index<1 then exit end if
      if index>len+plusOne then exit end if
      return index
   end while
   throw(Type_Check_Failure)
   return NIL
end function 
-- For these classes index 0 = length+1 (for insertions)
-- index -1 = end of sequence etc.
-- Note that the length of the sequence rather than the sequence itself is 
-- passed:  this is for efficiency, since your method will probably need 
-- to calculate the length of the sequence anyway.  
 
                           
package("Diamond Standard Class Library")
 
 
global constant Circular_Reference=
exception("Circular_Reference",Container_Error) 


class_entity Collection_class_handle 


global constant Component=
class("Component",ABSTRACT,Entity,Do_Not_Clone)

global type is_Component(object x)
   return live_instance_check(x,Component)
end type                           

property("Name",INSTANCE,PUBLIC,PRIVATE,PERSISTENT,NONE)
property("Collection",INSTANCE,PUBLIC,PRIVATE,PERSISTENT,Null_Instance)


method("new",0,CLASS,NORMAL,PROTECTED,NULL_METHOD,{Null_Instance})

function Component_new_1(is_identifier name)  
   entity new
   if not success() then return Null_Instance end if
   new=call_method(super(),"new",NONE)
   set_property(new,"Name",name)
   return new
end function  
  
method("new",1,CLASS,NORMAL,PROTECTED,routine_id("Component_new_1"),NONE)

function Component_delete_0()
   sequence collection    
   collection=get_property(this(),"Collection")
   if not equal(collection,Null_Instance) then
      set_property(this(),"Collection",Null_Instance)
      VOID=call_method(collection,"Remove",{get_property(this(),"Name")})
   end if
   return call_method(super(),"delete",NONE)
end function

method("delete",0,INSTANCE,NORMAL,PROTECTED,
routine_id("Component_delete_0"),NONE)     

function Component_Set_Collection_0()
   entity collection  
   collection=get_property(this(),"Collection")
   if equal(collection,Null_Instance) then 
      if extends(last(),Collection_class_handle) then             
         set_property(this(),"Collection",last())
         raise_event(this(),"Insert",NONE)
      end if                      
   elsif equal(collection,last()) then
      set_property(this(),"Collection",Null_Instance)
      raise_event(this(),"Remove",NONE)
   end if
   return NIL   
end function

method("Set_Collection",0,INSTANCE,FINAL,PUBLIC,
routine_id("Component_Set_Collection_0"),NONE)

event("Insert",0,INSTANCE,PROTECTED,PRIVATE)

event("Remove",0,INSTANCE,PROTECTED,PRIVATE) 

end_class()

                
global constant Collection=class("Collection",NORMAL,Component,Container)

global type is_Collection(object x)
   return live_instance_check(x,Collection)
end type                           

property("Components",INSTANCE,PRIVATE,PRIVATE,PERSISTENT,Null_Instance)
property("Component_Class",INSTANCE,PUBLIC,PRIVATE,PERSISTENT,Component)

function Collection_new_2(object name,object component)
   entity new                                     
   if not extends(component,Component) or deleted_instance(component) then
      throw(Type_Check_Failure)
      return Null_Instance   
   end if
   new=call_method(super(),"new",{name})
   if success() then 
      set_property(new,"Components",
      call_method(Indexed_Sequence,"new",{FALSE,FALSE,TRUE,1,FALSE}))
      set_property(new,"Component_Class",get_class(component))
      if instance_entity(component) then
         VOID=call_method(new,"Insert",{component})
      end if
   end if
   return new
end function  

method("new",1,CLASS,NORMAL,PUBLIC,
routine_id("Collection_new_2"),{Component})
method("new",2,CLASS,NORMAL,PUBLIC,routine_id("Collection_new_2"),NONE)

function Collection_delete_0()   
   for i=call_method(this(),"Count",NONE) to 1 by -1 do
      VOID=call_method(this(),"Remove",{i})
   end for
   VOID=call_method(get_property(this(),"Components"),"delete",NONE)                               
   return call_method(super(),"delete",NONE)
end function

method("delete",0,INSTANCE,NORMAL,PUBLIC,
routine_id("Collection_delete_0"),NONE)  

function Collection_Count_0()
   return call_method(get_property(this(),"Components"),"Count",NONE)
end function

method("Count",0,INSTANCE,FINAL,PUBLIC,routine_id("Collection_Count_0"),NONE)

function Collection_Get_1(object index)
   sequence component
   entity Components
   Components=get_property(this(),"Components")
   if integer(index) then 
      index=validate_index(index,call_method(Components,"Count",NONE),FALSE)
      if not success() then return Null_Instance end if
   else
      index=call_method(Components,"Find",{index})
      if index<1 then
         throw(Type_Check_Failure)
         return Null_Instance   
      end if
   end if
   component=call_method(Components,"Get",{index})      
   return component[1]
end function

method("Get",0,INSTANCE,FINAL,PUBLIC,routine_id("Collection_Get_1"),{-1})
method("Get",1,INSTANCE,FINAL,PUBLIC,routine_id("Collection_Get_1"),NONE)

function Collection_Find_1(is_identifier name)
   if not success() then return NIL end if
   return call_method(get_property(this(),"Components"),"Find",{name})
end function 

method("Find",1,INSTANCE,FINAL,PUBLIC,routine_id("Collection_Find_1"),NONE)

function Collection_Insert_1(is_Component component)
   object collection
   sequence name
   integer index
   entity Components                              
   if not success() then return NIL end if
   if not extends(component,get_property(this(),"Component_Class")) then
      throw(Type_Check_Failure)
      return NIL
   end if
   if extends(component,Collection) then
      collection=this()
      while TRUE do      
         index=find(collection,{Null_Instance,component})
         if index=1 then exit end if
         if index=2 then
            throw(Circular_Reference)
            return NIL
         end if
         collection=get_property(collection,"Collection")
      end while  
   end if
   name=get_property(component,"Name")
   Components=get_property(this(),"Components")
   VOID=call_method(Components,"Insert",{{name,component}})
   if success() then
      VOID=call_method(component,"Set_Collection",NONE)
   end if    
   return NIL
end function 

method("Insert",1,INSTANCE,FINAL,PUBLIC,
routine_id("Collection_Insert_1"),NONE) 

function Collection_Remove_1(object index)
   sequence component
   entity Components
   Components=get_property(this(),"Components")
   if integer(index) then
      index=validate_index(index,call_method(Components,"Count",NONE),FALSE)
      if not success() then return Null_Instance end if
   else
      index=call_method(Components,"Find",{index})
      if index<1 then 
         throw(Type_Check_Failure)
         return Null_Instance
      end if   
   end if                                             
   component=call_method(Components,"Remove",{index})
   component=component[1]
   if not equal(get_property(component,"Collection"),Null_Instance) then
      VOID=call_method(component,"Set_Collection",NONE)
   end if
   return component
end function  

method("Remove",0,INSTANCE,FINAL,PUBLIC,
routine_id("Collection_Remove_1"),{-1})
method("Remove",1,INSTANCE,FINAL,PUBLIC,
routine_id("Collection_Remove_1"),NONE)
  
end_class()

Collection_class_handle=Collection
  
  
package("Diamond Default")

